import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from setuptools.command.bdist_wheel import get_abi_tag, get_platform, tags


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        # Set the custom wheel tag
        build_data["tag"] = self._get_wheel_tag()
        print(f"Using custom wheel tag: {build_data['tag']}")
        # Run setup.py build_ext before main build
        subprocess.check_call(["python", "setup.py", "build_ext", "--inplace"])
        return super().initialize(version, build_data)

    def _get_wheel_tag(self):
        # Without the tag, the wheel will be named jcvi-0.0.0-py3-none-any.whl
        impl_name = tags.interpreter_name()
        impl_ver = tags.interpreter_version()
        abi_tag = get_abi_tag()
        plat_tag = get_platform(None)
        plat_tag = (  # macosx_11.0 => macosx_11_0
            plat_tag.lower().replace("-", "_").replace(".", "_").replace(" ", "_")
        )
        return f"{impl_name}{impl_ver}-{abi_tag}-{plat_tag}"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
%prog alignment.blocks alignment.bed layout.csv

Illustrate shared identity between aligned chromosomes. Use layout.csv to indicate
the positions of tracks. For example:

# x, y, rotation, ha, va, color, ratio, label, chrmName, rStart, rEnd, chrmMax
0.5, 0.6, 0, top, center, g, 1, Spp 1 Chrm 8, Chromosome_08, 2000000, 2108098, 2108098
0.5, 0.4, 0, right, center, m, 1, Spp 2 Chrm 1.4 , Scaffold_1.4, 50000, 200000, 2787645
0.5, 0.2, 45, bottom, center, b, 1, Spp3 Chr 5, SppX_Chromosome_05, 60000, 460000, 2046703
# edges
e, 0, 1
e, 1, 2

With the row ordering corresponding to the column ordering in the alignment blockfile.
"""

import sys
from typing import List

from matplotlib.path import Path
import numpy as np

from jcvi.apps.base import OptionParser, logger
from jcvi.compara.synteny import BlockFile
from jcvi.formats.bed import Bed
from jcvi.graphics.base import AbstractLayout, PathPatch, markup, mpl, plt, savefig
from jcvi.graphics.glyph import Glyph, RoundLabel  # , GeneGlyph
from jcvi.utils.cbook import human_size

# Default colours for ribbons with different orientations
forward, backward = "#1f77b4", "#2ca02c"


class LayoutLine(object):
    # 0-11: x,y,rotation,ha,va,color,ratio,label,chrmName,rStart,rEnd,chrmMax
    def __init__(self, row, delimiter=","):
        # Check if line is hidden (starts with '*')
        self.hidden = row[0] == "*"
        # If hidden, remove first character
        if self.hidden:
            row = row[1:]

        # Split row by delimiter
        args = row.rstrip().split(delimiter)
        # Strip whitespace from each arg
        args = [x.strip() for x in args]

        # Init layout line attributes
        self.ratio = 1
        self.label = None
        self.rStart = None
        self.rEnd = None
        self.chrmMax = None
        self.chrmName = None

        # Set layout line attributes
        self.x = float(args[0])
        self.y = float(args[1])
        self.rotation = int(args[2])  # rotation
        self.ha = args[3]  # horizontal alignment
        self.va = args[4]  # vertical alignment
        self.color = args[5]  # color
        if len(args) > 6 and args[6]:
            self.ratio = float(args[6])  # Scaling factor
        if len(args) > 7 and args[7]:
            self.label = args[7].strip()  # Chromosome label
        if len(args) > 8 and args[8]:
            self.chrmName = str(args[8])
        if len(args) > 9 and args[9] and args[10]:  # Set rStart and rEnd if provided
            # If rStart and rEnd are 0, set to 1
            self.rStart = max(int(args[9]), 1)
            self.rEnd = max(int(args[10]), 1)
        if len(args) > 10 and args[11]:
            self.chrmMax = int(args[11])


class Layout(AbstractLayout):
    def __init__(self, filename, delimiter=","):
        super(Layout, self).__init__(filename)  # Call parent class constructor
        fp = open(filename)
        self.edges = []
        for row in fp:
            # Skip blank lines
            if row.strip():
                # Skip header rows
                if row[0] == "#":
                    continue
                # Import edges
                if row[0] == "e":
                    # Split row by delimiter
                    args = row.rstrip().split(delimiter)
                    # Strip whitespace from each arg
                    args = [x.strip() for x in args]
                    # Check edge line format
                    assert args[0] == "e"
                    assert len(args) == 3
                    # From index 1 and 2 get source and target track indices
                    a, b = int(args[1]), int(args[2])
                    # Add edge to layout
                    self.edges.append((a, b))
                else:
                    # Parse all other lines as sequence tracks
                    self.append(LayoutLine(row, delimiter=delimiter))

        # Track number of layout lines added
        self.lines = len(self)

        if 2 <= len(self) <= 8:
            self.assign_colors()


class Shade(object):
    Styles = ("curve", "line")

    def __init__(
        self,
        ax,
        a,
        b,
        ymid,
        highlight=None,
        style="curve",
        ec="k",
        fc="k",
        alpha=0.2,
        lw=1,
        zorder=1,
    ):
        """Create syntenic wedges between tracks.
        Args:
            ax: matplotlib Axes
            # Note: a & b customised from synteny for ribbon module. Description is from synteny version.
            #a (tuple of floats): ((start_x, start_y), (end_x, end_y))
            #b (tuple of floats): ((start_x, start_y), (end_x, end_y))
            ymid (float): y-mid position for curve style only
            highlight (bool, optional): Plot this shade if color is specified. Defaults to False.
            style (str, optional): Style. Defaults to "curve", must be one of
            ("curve", "line")
            ec (str, optional): Edge color. Defaults to "k".
            fc (str, optional): Face color. Defaults to "k".
            alpha (float, optional): Transparency. Defaults to 0.2.
            lw (int, optional): Line width. Defaults to 1.
            zorder (int, optional): Z-order. Defaults to 1.
        """
        assert style in Shade.Styles, "style must be one of {}".format(self.Styles)

        # a1, a2 = a
        # b1, b2 = b
        a1, a2, a_start, a_end, a_strand, a_score = a
        b1, b2, b_start, b_end, b_strand, b_score = b
        ax1, ay1 = a1
        ax2, ay2 = a2
        bx1, by1 = b1
        bx2, by2 = b2
        M, C4, L, CP = Path.MOVETO, Path.CURVE4, Path.LINETO, Path.CLOSEPOLY
        if style == "curve":
            pathdata = [
                (M, a1),
                (C4, (ax1, ymid)),
                (C4, (bx1, ymid)),
                (C4, b1),
                (L, b2),
                (C4, (bx2, ymid)),
                (C4, (ax2, ymid)),
                (C4, a2),
                (CP, a1),
            ]
        else:
            pathdata = [(M, a1), (L, b1), (L, b2), (L, a2), (CP, a1)]
        codes, verts = list(zip(*pathdata))
        path = Path(verts, codes)

        if highlight and not ec:
            fc = highlight
        elif highlight:
            ec = fc = highlight

        pp = PathPatch(path, ec=ec, fc=fc, alpha=alpha, lw=lw, zorder=zorder)
        ax.add_patch(pp)


class Region(object):
    def __init__(
        self,
        ax,
        ext,
        layout,
        bed,
        scale,
        orientation=None,
        parent_ori=None,
        chr_label=True,
        loc_label=True,
        pad=0.05,
        vpad=0.015,
        features=None,
        extra_features=None,
        box_features=None,
        plotRibbonBlocks=False,
        annotcolor="g",
        extra_annotcolor="r",
    ):
        x, y = layout.x, layout.y
        ratio = layout.ratio
        scale /= ratio
        self.y = y
        lr = layout.rotation
        tr = mpl.transforms.Affine2D().rotate_deg_around(x, y, lr) + ax.transAxes
        inv = ax.transAxes.inverted()

        # Import standard track extent (by max feature range)
        start, end, si, ei, chrm, trackOri, span = ext
        startbp, endbp = (
            start.start,
            end.end,
        )  # Get start/end of first/last features from blockfile
        # Default all input tracks to Fwd orientation
        trackOri = "+"

        # Override track extent if custom range provided
        if layout.chrmName and layout.rStart and layout.rEnd:
            startbp, endbp, chrm = (layout.rStart, layout.rEnd, layout.chrmName)
            span = endbp - (startbp + 1)

        # Set flank size
        flank = span / scale / 2

        # Adjust x-coords to flank size
        xstart, xend = x - flank, x + flank
        self.xstart, self.xend = xstart, xend

        def cv(t):
            return xstart + abs(t - startbp) / scale

        hidden = layout.hidden

        # Plot Chromosome Bar
        if not hidden:
            ax.plot(
                (xstart, xend), (y, y), color="dimgray", transform=tr, lw=4, zorder=2
            )

        # Compose Chromosome label
        if orientation == "R" or trackOri == "-":
            label_startbp, label_endbp = endbp, startbp
        else:
            label_startbp, label_endbp = startbp, endbp

        if layout.label:
            chrm = layout.label

        label = "-".join(
            (
                human_size(label_startbp, target="Mb", precision=2)[:-2],
                human_size(label_endbp, target="Mb", precision=2),
            )
        )

        # Plot annotation tracks
        height = 0.012
        self.gg = {}

        # Process and plot ribbon intervals as annotation features
        self.ribbonBlocks = ribbonBlocks = bed[si : ei + 1]
        for g in ribbonBlocks:
            gstart, gend, outofrange = self.clip2range(g.start, g.end, startbp, endbp)
            if outofrange:
                continue
            # Set feature orientation - Note: Not really 'strand' just orientation relative to Parent where Parent is always '+'
            strand = g.strand
            if orientation == "R":
                gstart, gend = self.flip(gstart, gend, startbp, endbp)
                assert gstart <= gend
                # Flip feature strand (relative to parent) if self track has been flipped
                strand = "+" if strand == "-" else "-"
            # Invert gene start/end positions if feature on '-' strand
            if strand == "-":
                gstart, gend = gend, gstart
                assert gstart >= gend
            x1, x2, a, b = self.get_coordinates(gstart, gend, y, cv, tr, inv)
            # Note: gg is later used to plot ribbons
            # TODO: Create separate dict for (gstart, gend, strand, g.score) so that we don't have to tweak downstream functions that only want (a,b).
            self.gg[g.accn] = (a, b, gstart, gend, strand, g.score)
            # Filp feature orientation if Parent track has been flipped
            # TODO: Check if this is needs to happen before or after the gg update
            if parent_ori == "R":
                strand = "+" if strand == "-" else "-"
            color = forward if strand == "+" else backward
            if not hidden and plotRibbonBlocks:
                gp = Glyph(ax, x1, x2, y, height, gradient=False, fc=color, zorder=3)
                gp.set_transform(tr)

        # Set default feature colour
        feat_col = annotcolor.strip()

        # Set default feature colour for extra annotations
        extra_feat_col = extra_annotcolor.strip()

        # Add annotation track offset
        offset = 0.005

        # Set negative offset for extra annotations (below chromosome line)
        extra_offset = -offset

        # Set default annotation feature height
        feat_height = height * 0.3

        # Plot feature track (genes)
        if features:
            for g in features:
                # Bedline attributes =  seqid,start,end,accn,score,strand, extra = [color,offset,height]
                gstart, gend, outofrange = self.clip2range(
                    g.start, g.end, startbp, endbp
                )
                if outofrange:
                    continue
                if orientation == "R":
                    gstart, gend = self.flip(gstart, gend, startbp, endbp)
                x1, x2, a, b = self.get_coordinates(gstart, gend, y, cv, tr, inv)
                # Set custom annotation color / y-offset / height
                annot_col, annot_offset, annot_height = self.annotFormat(
                    g.extra, offset, feat_col, feat_height
                )
                # Note: Use GeneGlyph instead of Glyph for terminal CDS exon or for genes.
                # Need to store feature direction, and either flag terminal exons OR unique parent feature names (to determine last exon).
                # GeneGlyph(self, ax, x1, x2, y, height, gradient=True, tip=.0025, color="k", shadow=False)
                gp = Glyph(
                    ax,
                    x1,
                    x2,
                    y + annot_offset,
                    annot_height,
                    gradient=False,
                    fc=annot_col,
                    zorder=4,
                )
                gp.set_transform(tr)

        # Plot extra feature track - below chromosome
        if extra_features:
            for g in extra_features:
                gstart, gend, outofrange = self.clip2range(
                    g.start, g.end, startbp, endbp
                )
                if outofrange:
                    continue
                if orientation == "R":
                    gstart, gend = self.flip(gstart, gend, startbp, endbp)
                x1, x2, a, b = self.get_coordinates(gstart, gend, y, cv, tr, inv)
                # Set custom annotation color / y-offset / height
                annot_col, annot_offset, annot_height = self.annotFormat(
                    g.extra, extra_offset, extra_feat_col, feat_height
                )
                gp = Glyph(
                    ax,
                    x1,
                    x2,
                    y + annot_offset,  # Below chromosome with negative offset
                    annot_height,
                    gradient=False,
                    fc=annot_col,
                    zorder=4,
                )
                gp.set_transform(tr)

        if box_features:
            self.draw_annotation_boxes(
                ax, box_features, startbp, endbp, y, cv, tr, orientation
            )

        # Position and apply chromosome labels
        ha, va = layout.ha, layout.va

        hpad = 0.02
        if ha == "left":
            xx = xstart - hpad
            ha = "right"
        elif ha == "right":
            xx = xend + hpad
            ha = "left"
        else:
            xx = x
            ha = "center"

        # Tentative solution to labels stick into glyph
        magic = 40.0
        cc = abs(lr) / magic if abs(lr) > magic else 1
        if va == "top":
            yy = y + cc * pad
        elif va == "bottom":
            yy = y - cc * pad
        else:
            yy = y

        l = np.array((xx, yy))
        trans_angle = ax.transAxes.transform_angles(np.array((lr,)), l.reshape((1, 2)))[
            0
        ]
        lx, ly = l
        if not hidden:
            bbox = dict(boxstyle="round", fc="w", ec="w", alpha=0.5)
            kwargs = dict(
                ha=ha, va="center", rotation=trans_angle, bbox=bbox, zorder=10
            )

            # Format Chrm label for LaTeX
            chr_label = markup(chrm) if chr_label else None
            # Add seq coords to label
            loc_label = label if loc_label else None
            if chr_label:
                if loc_label:
                    ax.text(lx, ly + vpad, chr_label, color=layout.color, **kwargs)
                    ax.text(
                        lx,
                        ly - vpad,
                        loc_label,
                        color="lightslategrey",
                        size=10,
                        **kwargs,
                    )
                else:
                    ax.text(lx, ly, chr_label, color=layout.color, **kwargs)

    def annotFormat(self, extra, offset, feat_col, feat_height):
        annot_col = feat_col
        annot_offset = offset
        annot_height = feat_height
        # extra [colour,offset multiplier, height multiplier]
        if extra:
            if len(extra) >= 1 and str(extra[0]) != ".":
                annot_col = str(extra[0])
            if len(extra) >= 2 and str(extra[1]) != ".":
                annot_offset = offset * float(extra[1])
            if len(extra) >= 3 and str(extra[2]) != "." and float(extra[2]) > 0:
                annot_height = feat_height * float(extra[2])
        return (annot_col, annot_offset, annot_height)

    def get_coordinates(self, gstart, gend, y, cv, tr, inv):
        x1, x2 = cv(gstart), cv(gend)
        a, b = tr.transform((x1, y)), tr.transform((x2, y))
        a, b = inv.transform(a), inv.transform(b)
        return x1, x2, a, b

    def clip2range(self, f_start, f_end, start, end):
        # Return unchanged if annotation within plot range
        if f_start >= start and f_end <= end:
            return f_start, f_end, False
        # Clip if plot range fully within annotation
        if f_start <= start and f_end >= end:
            return start, end, False
        # If annotation starts before plot range
        if f_start <= start:
            # And ends before plot range, exclude annotation
            if f_end <= start:
                return None, None, True
            # If end is within plot range, clip to range start
            elif f_end >= start:
                return start, f_end, False
        # If annotation ends after plot range
        if f_end >= end:
            # And begins after range end, exclude annotation
            if f_start >= end:
                return None, None, True
            # If annotation start is within plot range, clip to range end
            elif f_start <= end:
                return f_start, end, False

    def flip(self, gstart, gend, startbp, endbp):
        upstream = gstart - startbp
        downstream = endbp - gend
        newEnd = endbp - upstream
        newStart = startbp + downstream
        return newStart, newEnd

    def draw_annotation_boxes(
        self, ax, box_features, startbp, endbp, y, cv, tr, orientation
    ):
        # Same offset used for annotations
        offset = 0.005
        feat_height = 0.012 * 0.3
        box_padding = 0.001
        box_height = 2 * (offset + feat_height + box_padding)

        # Border width in points
        lw = 2

        # Get figure for coordinate conversion
        fig = ax.figure

        # Calculate HALF the line width in figure coordinates
        # Line width is centered on the path, so we need half to move borders out precisely
        fig_width_inches = fig.get_figwidth()
        half_lw_fig_units = (lw / 2) / (72 * fig_width_inches)

        for g in box_features:
            # Clip box to chromosome range
            gstart, gend, outofrange = self.clip2range(g.start, g.end, startbp, endbp)
            if outofrange:
                continue

            # Apply orientation flipping if needed
            if orientation == "R":
                gstart, gend = self.flip(gstart, gend, startbp, endbp)

            # Get horizontal coordinates
            x1 = cv(gstart)
            x2 = cv(gend)

            # Adjust coordinates to move borders outward by HALF the line width
            x1 -= half_lw_fig_units  # Move left border out
            x2 += half_lw_fig_units  # Move right border out

            # Get box color from BED file
            box_color = "black"
            if g.extra and len(g.extra) >= 1 and g.extra[0] != ".":
                box_color = g.extra[0]

            # Create box with adjusted coordinates
            rect = mpl.patches.Rectangle(
                (x1, y - box_height / 2),
                x2 - x1,
                box_height,
                linewidth=lw,
                edgecolor=box_color,
                facecolor="none",
                alpha=0.8,
                zorder=5,
            )
            rect.set_transform(tr)
            ax.add_patch(rect)


class Synteny(object):
    def __init__(
        self,
        fig,
        root,
        blocks,
        bedfile,
        layoutfile,
        orientation=None,
        tree=None,
        features=None,
        extra_features=None,
        box_annotations=None,
        chr_label=True,
        loc_label=True,
        pad=0.05,
        vpad=0.015,
        scalebar=False,
        paintbyscore=False,
        paintbystrand=False,
        prune_features=True,
        plotRibbonBlocks=False,
        annotcolor="g",
        extra_annotcolor="r",
        scaleAlpha=False,
        noline=False,
        shadestyle="curve",
    ):
        # Set figure dimensions
        w, h = fig.get_figwidth(), fig.get_figheight()
        # Import primary annotations from bedfile
        bed = Bed(bedfile)
        # Order bed features as dict {'ppa004748m': (42353, scaffold_5    1666610 1670773 ppa004748m  0   +),..}
        # genename: (orderedPosition, seqname, start, end, genename, score, strand)
        order = bed.order
        # Import synteny links (pairs of source > target IDs for features in BED file)
        bf = BlockFile(blocks, defaultcolor="hide", allow_hex=True)
        # Import layout config
        self.layout = lo = Layout(layoutfile)

        # Check correct number of orientation values
        if orientation:
            logger.debug(f"Set track orientations {orientation}")
            if lo.lines != len(orientation):
                logger.error("Incorrect number of orientation instructions")
                sys.exit(0)

        # Import features track
        if features:
            features = Bed(features)

        # Import extra features track
        if extra_features:
            extra_features = Bed(extra_features)

        # Init list for track extent, features, and extra features
        exts = []
        feats = []
        extra_feats = []
        box_feats = []

        # Collect span for any tracks with custom range coords
        customSpans = []

        # For each track (by index)
        logger.debug(f"bf.ncols is {bf.ncols}")

        for i in range(bf.ncols):
            # Set range coord containing all annotations
            ext = bf.get_extent(i, order)
            # Add to ext index
            exts.append(ext)
            # ext has format:
            # (  chr1  10730   27033   GSVIVT01012261001   0   +,
            #   chr1 558160  562819  GSVIVT01012208001   0   +,
            #   0,
            #   49,
            #   'chr1',
            #   '+',
            #   552088)
            # = first feature, last feature, bf idx ff, bf idx lf, chr name,
            #   track orientation, span (range from ff to lf)

            # Override block feature extents if custom range provided in layout
            if lo[i].chrmName and lo[i].rStart and lo[i].rEnd:
                start, end, chrm = (lo[i].rStart, lo[i].rEnd, lo[i].chrmName)
                span = end - (start + 1)
                customSpans.append(span)
                logger.debug(
                    f"start is {start}, end is {end}, chrm is {chrm}, orientation is {orientation[i]}, span is {span}"
                )
            else:
                # TODO: Check what ori is doing here
                start, end, si, ei, chrm, ori, span = ext
                # Get start/end of first/last features from blockfile
                start, end = start.start, end.end
                customSpans.append(span)

                logger.debug(
                    f"start is {start}, end is {end}, si is {si}, ei is {ei}, chrm is {chrm}, orientation is {orientation[i]}, span is {span}"
                )

            if features:
                # Unpack coords from 'features' bed object
                # Filter for features within range to be plotted
                fe = list(features.extract(chrm, start, end))
                # Pruning removes minor features with < 0.1% of the region
                if prune_features:
                    fe_pruned = [x for x in fe if x.span >= span / 1000]
                    logger.info(
                        "Extracted {0} features "
                        "({1} after pruning)".format(len(fe), len(fe_pruned)),
                        file=sys.stderr,
                    )
                    feats.append(fe_pruned)
                else:
                    fe_all = [x for x in fe]
                    feats.append(fe_all)

            # Process extra annotations
            if extra_features:
                # Extract features for this chromosome
                extra_fe = list(extra_features.extract(chrm, start, end))
                if prune_features:
                    extra_fe_pruned = [x for x in extra_fe if x.span >= span / 1000]
                    logger.info(
                        "Extracted {0} extra features "
                        "({1} after pruning)".format(
                            len(extra_fe), len(extra_fe_pruned)
                        ),
                        file=sys.stderr,
                    )
                    extra_feats.append(extra_fe_pruned)
                else:
                    extra_fe_all = [x for x in extra_fe]
                    extra_feats.append(extra_fe_all)
            else:
                # Add empty list if no extra features
                extra_feats.append([])

        if box_annotations:
            box_annotations = Bed(box_annotations)
            for i in range(bf.ncols):
                # Get chromosome-specific information
                if lo[i].chrmName and lo[i].rStart and lo[i].rEnd:
                    track_chrm, track_start, track_end = (
                        lo[i].chrmName,
                        lo[i].rStart,
                        lo[i].rEnd,
                    )
                else:
                    _, _, _, _, track_chrm, _, _ = exts[i]
                    track_start, track_end = exts[i][0].start, exts[i][1].end

                # Extract relevant box annotations for this chromosome
                box_fe = list(
                    box_annotations.extract(track_chrm, track_start, track_end)
                )
                box_feats.append(box_fe)

        # Find largest coord range for any track
        # maxspan = max(exts, key=lambda x: x[-1])[-1]
        maxspan = max(customSpans)
        scale = maxspan / 0.65

        self.gg = gg = {}
        self.rr = []
        ymids = []

        # Plot annotations
        for i in range(bf.ncols):
            logger.debug(f"annotation {i} orientation is {orientation[i]}")
            ori = orientation[i] if orientation else None
            parent_ori = orientation[i - 1] if orientation and i >= 1 else None
            r = Region(
                root,
                exts[i],
                lo[i],
                bed,
                scale,
                orientation=ori,
                parent_ori=parent_ori,
                chr_label=chr_label,
                loc_label=loc_label,
                vpad=vpad,
                features=feats[i] if feats else None,
                extra_features=(
                    extra_feats[i] if extra_feats else None
                ),  # Pass extra features to Region
                box_features=box_feats[i] if box_annotations else None,
                plotRibbonBlocks=plotRibbonBlocks,
                annotcolor=annotcolor,
                extra_annotcolor=extra_annotcolor,  # Pass extra color to Region
            )
            self.rr.append(r)
            # Use tid and accn to store gene positions
            gg.update(dict(((i, k), v) for k, v in list(r.gg.items())))
            ymids.append(r.y)

        # Plot ribbons
        for i, j in lo.edges:
            # geneA, geneB, highlight colour
            for ga, gb, h in bf.iter_pairs(i, j):
                # Coords in gg should be correctly trimmed from Region processing
                # Need to skip annotations in blockfile that are outside plotted region in layout file
                if (i, ga) not in list(gg.keys()) or (j, gb) not in list(gg.keys()):
                    continue
                # Skip ribbons that will later be plotted with a hightlight color
                if h:
                    continue
                a, b = gg[(i, ga)], gg[(j, gb)]
                ymid = (ymids[i] + ymids[j]) / 2
                # Draw synteny ribbons
                ribbonColor = "gainsboro"
                baseAlpha = 0.8
                if paintbystrand:
                    ribbonColor = self.inversionCheck(a, b)
                if paintbyscore:
                    baseAlpha = self.scoreCheck(a, scaleAlpha)

                if noline:
                    lw = 0.0
                    edge_color = None
                else:
                    lw = 0.1
                    edge_color = ribbonColor
                Shade(
                    root,
                    a,
                    b,
                    ymid,
                    fc=ribbonColor,
                    ec=edge_color,
                    lw=lw,
                    alpha=baseAlpha,
                    style=shadestyle,
                )
                # Add second low alpha ribbon track with thin outlines. Helps visualise very thin bands.
                # Shade(root, a, b, ymid, fc=ribbonColor, lw=0.1, alpha=0.125, style=shadestyle)

            # Paint ribbons for which a highlight colour was set
            for ga, gb, h in bf.iter_pairs(i, j, highlight=True):
                if (i, ga) not in list(gg.keys()) or (j, gb) not in list(gg.keys()):
                    continue
                if h == "hide":
                    continue
                a, b = gg[(i, ga)], gg[(j, gb)]
                ymid = (ymids[i] + ymids[j]) / 2
                baseAlpha = 1
                if paintbyscore:
                    baseAlpha = self.scoreCheck(a, scaleAlpha)

                if noline:
                    lw = 0.0
                    edge_color = None
                else:
                    lw = 0.1
                    edge_color = h
                Shade(
                    root,
                    a,
                    b,
                    ymid,
                    alpha=baseAlpha,
                    highlight=h,
                    ec=edge_color,
                    lw=lw,
                    zorder=1,
                    style=shadestyle,
                )

        if scalebar:
            logger.info("Build scalebar (scale={})".format(scale), file=sys.stderr)
            # Find the best length of the scalebar
            ar = [1, 2, 5]
            candidates = (
                [1000 * x for x in ar]
                + [10000 * x for x in ar]
                + [100000 * x for x in ar]
            )
            # Find the one that's close to an optimal canvas size
            dists = [(abs(x / scale - 0.12), x) for x in candidates]
            dist, candidate = min(dists)
            dist = candidate / scale
            x, y, yp = 0.2, 0.96, 0.005
            a, b = x - dist / 2, x + dist / 2
            lsg = "lightslategrey"
            root.plot([a, a], [y - yp, y + yp], "-", lw=2, color=lsg)
            root.plot([b, b], [y - yp, y + yp], "-", lw=2, color=lsg)
            root.plot([a, b], [y, y], "-", lw=2, color=lsg)
            root.text(
                x,
                y + 0.02,
                human_size(candidate, precision=0),
                ha="center",
                va="center",
            )

        if tree:
            from jcvi.graphics.tree import draw_tree, read_trees

            trees = read_trees(tree)
            ntrees = len(trees)
            logger.debug("A total of {0} trees imported.".format(ntrees))
            xiv = 1.0 / ntrees
            yiv = 0.3
            xstart = 0
            ystart = min(ymids) - 0.4
            for i in range(ntrees):
                ax = fig.add_axes([xstart, ystart, xiv, yiv])
                label, outgroup, tx = trees[i]
                draw_tree(ax, tx, outgroup=outgroup, rmargin=0.4, leaffont=11)
                xstart += xiv
                RoundLabel(ax, 0.5, 0.3, label, fill=True, fc="lavender", color="r")

    def inversionCheck(self, a, b):
        a1, a2, a_start, a_end, a_strand, a_score = a
        b1, b2, b_start, b_end, b_strand, b_score = b
        aDir = "F" if a_start < a_end else "R"
        bDir = "F" if b_start < b_end else "R"
        if aDir == bDir:
            return forward
        else:
            return backward

    def scoreCheck(self, a, scaleAlpha):
        a1, a2, a_start, a_end, a_strand, a_score = a
        assert 0 <= float(a_score) <= 100
        if scaleAlpha:
            OldValue = float(a_score)
            OldMin = 50
            OldMax = 100
            NewMax = 90
            NewMin = 10
            newScore = (
                ((OldValue - OldMin) * (NewMax - NewMin)) / (OldMax - OldMin)
            ) + NewMin
            if newScore < NewMin:
                newScore = NewMin
            return float(newScore) / 100
        else:
            return float(a_score) / 100


def draw_gene_legend(ax, x1, x2, ytop, d=0.04, text=False, repeat=False):
    ax.plot([x1, x1 + d], [ytop, ytop], ":", color=forward, lw=2)
    ax.plot([x1 + d], [ytop], ">", color=forward, mec=forward)
    ax.plot([x2, x2 + d], [ytop, ytop], ":", color=backward, lw=2)
    ax.plot([x2], [ytop], "<", color=backward, mec="g")
    if text:
        ax.text(x1 + d / 2, ytop + d / 2, "gene (+)", ha="center")
        ax.text(x2 + d / 2, ytop + d / 2, "gene (-)", ha="center")
    if repeat:
        xr = (x1 + x2 + d) / 2
        Glyph(
            ax,
            xr - d / 2,
            xr + d / 2,
            ytop,
            0.012 * 3 / 4,
            gradient=False,
            fc="#ff7f00",
            zorder=2,
        )
        ax.text(xr, ytop + d / 2, "repeat", ha="center")


def set_strand_colors(colorCodes):
    global forward
    global backward
    forward, backward = colorCodes.strip().split(",")


def main(args: List[str]):
    # Get cmd line args
    p = OptionParser(__doc__)
    p.add_argument("--outfile", default=None, help="Prefix for output graphic.")
    p.add_argument(
        "--annotations",
        help="""
        Feature annotations in BED format. \n \
        [1] seqid \n \
        [2] start \n \
        [3] end \n \
        [4] accn \n \
        [5] score \n \
        [6] strand \n \
        [7] Custom color. i.e. '.' = use default color, else any pyplot compatible color code: 'g', 'green', '#02ab2e', etc. \n \
        [8] Vertical offset multiplier. i.e. 0 = plot on chrm line, -1 = plot below chrm, 1 = plot above chrm \n \
        [9] Feature height multiplier. i.e. 1 = default, 2 = double height
        """,
    )
    p.add_argument(
        "--extra-annotations",
        default=None,
        help="""Extra annotations in BED format. \n \
        [1] seqid \n \
        [2] start \n \
        [3] end \n \
        [4] accn \n \
        [5] score \n \
        [6] strand \n \
        [7] Custom color. i.e. '.' = use default color, else any pyplot compatible color code: 'g', 'green', '#02ab2e', etc. \n \
        [8] Vertical offset multiplier. i.e. 0 = plot on chrm line, -1 = plot below chrm, 1 = plot above chrm \n \
        [9] Feature height multiplier. i.e. 1 = default, 2 = double height """,
    )
    p.add_argument(
        "--annotation-boxes",
        default=None,
        help="Draw boxes with colored borders around chromosome regions. BED format: "
        "chr, start, end, name, score, strand, color. Boxes extend to include "
        "annotation tracks.",
    )
    p.add_argument(
        "--noprune",
        default=True,
        action="store_false",
        help="If set, do not exclude small features from annotation track. ",
    )
    p.add_argument(
        "--reorient",
        default=None,
        help="Comma delimited string of 'F' or 'R' characters. Must be same number and order as tracks in layout file. \n i.e. F,F,R will flip the third track. Default: All Forward.",
    )
    p.add_argument("--tree", help="Display trees on the bottom of the figure.")
    p.add_argument(
        "--scalebar",
        default=False,
        action="store_true",
        help="Add scale bar to the plot",
    )
    p.add_argument(
        "--paintbyscore",
        default=False,
        action="store_true",
        help="Set ribbon transparancy using score column from ribbon bedfile.",
    )
    p.add_argument(
        "--paintbystrand",
        default=False,
        action="store_true",
        help="Set ribbon colour by alignment orientation. Red=Inverted, Blue=Same",
    )
    p.add_argument(
        "--plotHits",
        default=False,
        action="store_true",
        help="If set, plot ribbon features from blockfile also as annotations.",
    )
    p.add_argument(
        "--strandcolors",
        default=None,
        help="Comma delimited string of color codes for forward or inverted orientation ribbons. Used by paintbystrand. Default: 'b,g' for forward,inverted.",
    )
    p.add_argument(
        "--annotcolor",
        default="g",
        help="Comma delimited string of color code for 'annotation' feature tracks. Default: 'g'.",
    )
    p.add_argument(
        "--scaleAlpha",
        default=False,
        action="store_true",
        help="If set, ribbon alpha values will be rescaled from 0.5-1 to 0.15-0.95 Note: Assumes 50%% min identity in alignments.",
    )
    p.add_argument(
        "--transparent",
        default=False,
        action="store_true",
        help="If set, save image with transparent background.",
    )
    p.add_argument(
        "--noline",
        default=False,
        action="store_true",
        help="If set, do not draw outline on ribbons.",
    )
    p.add_argument(
        "--shadestyle",
        default="curve",
        choices=Shade.Styles,
        help="Style of syntenic wedges",
    )
    # Unpack options
    # opts = {'style': 'darkgrid', 'format': 'pdf', 'scalebar': False, 'extra': 'grape_peach_cacao.bed', 'tree': None, 'diverge': 'PiYG', 'cmap': 'jet', 'figsize': '8x7', 'font': 'Helvetica', 'dpi': 300}
    # args = positional args, data files
    # iopts = (2400px x 2100px)
    opts, args, iopts = p.set_image_options(args, figsize="20x20")

    # Check for data files
    if len(args) != 3:
        logger.error(f"Positional args should have 3 items: {args}")
        sys.exit(not p.print_help())

    # Unpack data file paths
    blocks, bedfile, layoutfile = args

    # Set strand colors as global vars
    if opts.strandcolors:
        set_strand_colors(opts.strandcolors)

    # Get custom track orientations
    # TODO: raise error if orientation string is not same length as number of tracks
    # TODO: raise error if orientation string contains anything other than 'F' or 'R'
    if opts.reorient:
        flip = [x for x in opts.reorient.strip().split(",") if x in ("F", "R")]
    else:
        flip = None

    # Create base image
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    # Plot all the things!
    Synteny(
        fig,
        root,
        blocks,
        bedfile,
        layoutfile,
        orientation=flip,
        tree=opts.tree,
        features=opts.annotations,
        extra_features=opts.extra_annotations,
        box_annotations=opts.annotation_boxes,
        scalebar=opts.scalebar,
        paintbyscore=opts.paintbyscore,
        paintbystrand=opts.paintbystrand,
        prune_features=opts.noprune,
        plotRibbonBlocks=opts.plotHits,
        annotcolor=opts.annotcolor,
        scaleAlpha=opts.scaleAlpha,
        noline=opts.noline,
        shadestyle=opts.shadestyle,
    )

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    if opts.outfile:
        image_name = str(opts.outfile) + "." + iopts.format
    else:
        pf = blocks.rsplit(".", 1)[0]
        image_name = pf + "." + iopts.format

    savefig(image_name, dpi=iopts.dpi, iopts=iopts, transparent=opts.transparent)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import copy
from functools import partial
from os import remove
import os.path as op
import re
import sys
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import seaborn as sns

mpl.use("Agg")

from brewer2mpl import get_map
from matplotlib import cm, rc, rcParams
import matplotlib.colors as mcolors
from matplotlib.colors import Colormap
from matplotlib.patches import (
    CirclePolygon,
    Ellipse,
    FancyArrow,
    FancyArrowPatch,
    FancyBboxPatch,
    PathPatch,
    Polygon,
    Rectangle,
)
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

from ..apps.base import datadir, glob, logger, sample_N, which
from ..formats.base import LineFile
from ..utils.cbook import human_size

Extent = Tuple[float, float, float, float]

CHARS = {
    "&": r"\&",
    "%": r"\%",
    "$": r"\$",
    "#": r"\#",
    "_": r"\_",
    "{": r"\{",
    "}": r"\}",
}

GRAPHIC_FORMATS = (
    "emf",
    "eps",
    "pdf",
    "png",
    "ps",
    "raw",
    "rgba",
    "svg",
    "svgz",
)


def is_tex_available() -> bool:
    """Check if latex command is available"""
    return bool(which("latex")) and bool(which("lp"))


class ImageOptions(object):
    def __init__(self, opts):
        self.w, self.h = [int(x) for x in opts.figsize.split("x")]
        self.dpi = opts.dpi
        self.format = opts.format
        self.cmap = mpl.colormaps[opts.cmap]
        self.seed = opts.seed
        self.usetex = is_tex_available() and not opts.notex
        self.opts = opts

    def __str__(self):
        return "({0}px x {1}px)".format(self.dpi * self.w, self.dpi * self.h)

    @property
    def diverge(self):
        colors = get_map(self.opts.diverge, "diverging", 5).mpl_colors
        return colors[0], colors[-1]


class TextHandler(object):
    def __init__(self, fig, usetex: bool = True):
        self.heights = []
        try:
            self.build_height_array(fig, usetex=usetex)
        except ValueError as e:
            logger.debug(
                "Failed to init heights (error: %s). Variable label sizes skipped.", e
            )

    @classmethod
    def get_text_width_height(cls, fig, txt="chr01", size=12, usetex: bool = True):
        tp = mpl.textpath.TextPath((0, 0), txt, size=size, usetex=usetex)
        bb = tp.get_extents()
        xmin, ymin = fig.transFigure.inverted().transform((bb.xmin, bb.ymin))
        xmax, ymax = fig.transFigure.inverted().transform((bb.xmax, bb.ymax))
        return xmax - xmin, ymax - ymin

    def build_height_array(self, fig, start=1, stop=36, usetex: bool = True):
        for i in range(start, stop + 1):
            w, h = TextHandler.get_text_width_height(fig, size=i, usetex=usetex)
            self.heights.append((h, i))

    def select_fontsize(self, height, minsize=1, maxsize=12):
        if not self.heights:
            return maxsize if height > 0.01 else minsize

        from bisect import bisect_left

        i = bisect_left(self.heights, (height,))
        size = self.heights[i - 1][1] if i else minsize
        size = min(size, maxsize)
        return size


class AbstractLayout(LineFile):
    """
    Simple csv layout file for complex plotting settings. Typically, each line
    represents a subplot, a track or a panel.
    """

    def __init__(self, filename):
        super().__init__(filename)

    def assign_array(self, attrib, array):
        assert len(array) == len(self)
        for x, c in zip(self, array):
            if not getattr(x, attrib):
                setattr(x, attrib, c)

    def assign_colors(self, seed: Optional[int] = None):
        number = len(self)
        palette = set2_n if number <= 8 else set3_n
        # Restrict palette numbers between [3, 12]
        palette_number = max(3, min(number, 12))
        colorset = palette(palette_number)
        colorset = sample_N(colorset, number, seed=seed)
        self.assign_array("color", colorset)

    def assign_markers(self, seed: Optional[int] = None):
        markerset = sample_N(mpl.lines.Line2D.filled_markers, len(self), seed=seed)
        self.assign_array("marker", markerset)

    def __str__(self):
        return "\n".join(str(x) for x in self)


def adjust_extent(extent: Extent, root_extent: Extent) -> Extent:
    """
    Adjust the extent of the root axes.
    """
    rx, ry, rw, rh = root_extent
    ex, ey, ew, eh = extent
    return rx + ex * rw, ry + ey * rh, ew * rw, eh * rh


def linear_blend(from_color, to_color, fraction=0.5):
    """Interpolate a new color between two colors.

    https://github.com/PimpTrizkit/PJs/wiki/12.-Shade,-Blend-and-Convert-a-Web-Color-(pSBC.js)

    Args:
        from_color (matplotlib color): starting color
        to_color (matplotlib color): ending color
        fraction (float, optional): Range is 0 (closer to starting color) to 1
        (closer to ending color). Defaults to 0.5.
    """
    from matplotlib.colors import to_rgb

    def lerp(v0, v1, t):
        # Precise method, which guarantees v = v1 when t = 1
        return (1 - t) * v0 + t * v1

    r1, g1, b1 = to_rgb(from_color)
    r2, g2, b2 = to_rgb(to_color)
    return lerp(r1, r2, fraction), lerp(g1, g2, fraction), lerp(b1, b2, fraction)


def linear_shade(from_color, fraction=0.5):
    """Interpolate a lighter or darker color.

    https://github.com/PimpTrizkit/PJs/wiki/12.-Shade,-Blend-and-Convert-a-Web-Color-(pSBC.js)

    Args:
        from_color (matplotlib color): starting color
        fraction (float, optional): Range is -1 (darker) to 1 (lighter). Defaults to 0.5.
    """
    assert -1 <= fraction <= 1, "Fraction must be between -1 and 1"
    if fraction < 0:
        return linear_blend("k", from_color, 1 + fraction)
    return linear_blend(from_color, "w", fraction)


def load_image(filename: str) -> np.ndarray:
    """
    Load an image file and return as numpy array.
    """
    img = plt.imread(filename)
    if len(img.shape) == 2:  # Gray-scale image, convert to RGB
        # http://www.socouldanyone.com/2013/03/converting-grayscale-to-rgb-with-numpy.html
        h, w = img.shape
        ret = np.empty((h, w, 3), dtype=np.uint8)
        ret[:, :, 2] = ret[:, :, 1] = ret[:, :, 0] = img
        img = ret
    else:
        h, w, _ = img.shape
    logger.debug("Image `%s` loaded (%dpx x %dpx).", filename, w, h)
    return img


def latex(s):
    """Latex doesn't work well with certain characters, like '_', in plain text.
    These characters would be interpreted as control characters, so we sanitize
    these strings.

    Args:
        s (str): Input string

    Returns:
        str: Output string sanitized
    """
    return "".join([CHARS.get(char, char) for char in s])


def shorten(s, maxchar=20, mid="..."):
    if len(s) <= maxchar or len(mid) >= maxchar:
        return s
    pad = (maxchar - len(mid)) // 2
    right_pad = maxchar - len(mid) - pad
    return s[:pad] + mid + s[-right_pad:]


def set1_n(number=9):
    return get_map("Set1", "qualitative", number).hex_colors


def set2_n(number=8):
    # Get Set2 from ColorBrewer, a set of colors deemed colorblind-safe and
    # pleasant to look at by Drs. Cynthia Brewer and Mark Harrower of Pennsylvania
    # State University. These colors look lovely together, and are less
    # saturated than those colors in Set1.
    return get_map("Set2", "qualitative", number).hex_colors


def set3_n(number=12):
    return get_map("Set3", "qualitative", number).hex_colors


def paired_n(number=12):
    """See also: https://colorbrewer2.org/#type=qualitative&scheme=Paired&n=12"""
    return get_map("Paired", "qualitative", number).hex_colors


set1, set2, set3, paired = set1_n(), set2_n(), set3_n(), paired_n()


def prettyplot():
    reds = copy.copy(mpl.cm.Reds)
    reds.set_bad("white")
    reds.set_under("white")

    blues_r = copy.copy(mpl.cm.Blues_r)
    blues_r.set_bad("white")
    blues_r.set_under("white")

    # Need to 'reverse' red to blue so that blue=cold=small numbers,
    # and red=hot=large numbers with '_r' suffix
    blue_red = get_map("RdBu", "diverging", 11, reverse=True).mpl_colormap
    green_purple = get_map("PRGn", "diverging", 11).mpl_colormap
    red_purple = get_map("RdPu", "sequential", 9).mpl_colormap

    return blues_r, reds, blue_red, green_purple, red_purple


blues_r, reds, blue_red, green_purple, red_purple = prettyplot()


def normalize_axes(*axes):
    """
    Normalize the axes to have the same scale.
    """
    for ax in axes:
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        ax.set_axis_off()


def panel_labels(ax, labels, size: int = 16):
    """
    Add panel labels (A, B, ...) to a figure.
    """
    for xx, yy, panel_label in labels:
        if rcParams["text.usetex"]:
            panel_label = r"$\textbf{{{0}}}$".format(panel_label)
        ax.text(xx, yy, panel_label, size=size, ha="center", va="center")


def update_figname(figname: str, format: str) -> str:
    """Update the name of a figure to include the format.

    Args:
        figname (str): Path to the figure
        format (str): Figure format, must be one of GRAPHIC_FORMATS

    Returns:
        str: New file path
    """
    _, ext = op.splitext(figname)
    if ext.strip(".") in GRAPHIC_FORMATS:  # User suffix has precedence
        return figname
    # When the user has not supplied a format in the filename, use the requested format
    assert format in GRAPHIC_FORMATS, "Invalid format"
    return figname + "." + format


def savefig(figname, dpi=150, iopts=None, cleanup=True, transparent=False):
    try:
        format = figname.rsplit(".", 1)[-1].lower()
    except:
        format = "pdf"
    logger.debug("Matplotlib backend is: %s", mpl.get_backend())
    try:
        logger.debug("Attempting save as: `%s`", figname)
        plt.savefig(figname, dpi=dpi, format=format, transparent=transparent)
    except Exception as e:
        message = "savefig failed with message:"
        message += "\n{0}".format(str(e))
        logger.error(message)
        logger.debug("Attempted save as: %s", format)
        logger.info("Try running again with --notex option to disable latex.")
        if op.exists(figname):
            if op.getsize(figname) < 1000:
                logger.debug("Cleaning up empty file: %s", figname)
                remove(figname)
        sys.exit(1)

    msg = f"Figure saved to `{figname}`"
    if iopts:
        msg += f" {iopts}"
    logger.debug(msg)

    if cleanup:
        plt.rcdefaults()


# human readable size (Kb, Mb, Gb)
def human_readable(x: Union[str, int], _, base=False):
    x = str(int(x))
    if x.endswith("000000000"):
        x = x[:-9] + "G"
    elif x.endswith("000000"):
        x = x[:-6] + "M"
    elif x.endswith("000"):
        x = x[:-3] + "K"
    if base and x[-1] in "MKG":
        x += "b"
    return x


human_readable_base = partial(human_readable, base=True)
human_formatter = ticker.FuncFormatter(human_readable)
human_base_formatter = ticker.FuncFormatter(human_readable_base)
mb_formatter = ticker.FuncFormatter(lambda x, pos: "{0}M".format(int(x / 1000000)))
mb_float_formatter = ticker.FuncFormatter(
    lambda x, pos: "{0:.1f}M".format(x / 1000000.0)
)
kb_formatter = ticker.FuncFormatter(lambda x, pos: "{0}K".format(int(x / 1000)))


def set_human_axis(ax, formatter=human_formatter):
    ax.xaxis.set_major_formatter(formatter)
    ax.yaxis.set_major_formatter(formatter)


set_human_base_axis = partial(set_human_axis, formatter=human_base_formatter)


def set_helvetica_axis(ax):
    xtick_locs = ax.get_xticks().tolist()
    ytick_locs = ax.get_yticks().tolist()
    # If we dont do the following, we have
    # UserWarning: FixedFormatter should only be used together with FixedLocator
    ax.xaxis.set_major_locator(mpl.ticker.FixedLocator(xtick_locs))
    ax.yaxis.set_major_locator(mpl.ticker.FixedLocator(ytick_locs))
    ax.set_xticklabels([int(x) for x in xtick_locs], family="Helvetica")
    ax.set_yticklabels([int(x) for x in ytick_locs], family="Helvetica")


available_fonts = [op.basename(x) for x in glob(datadir + "/*.ttf")]


def fontprop(ax, name, size=12):
    assert name in available_fonts, "Font must be one of {0}.".format(available_fonts)

    import matplotlib.font_manager as fm

    fname = op.join(datadir, name)
    prop = fm.FontProperties(fname=fname, size=size)

    logger.debug("Set font to `%s` (`%s`)", name, prop.get_file())
    for text in ax.texts:
        text.set_fontproperties(prop)

    return prop


def markup(s: str):
    """
    Change the string to latex format, and italicize the text between *.
    """
    if not rcParams["text.usetex"]:
        return s
    if "$" in s:
        return s
    s = latex(s)
    s = re.sub(r"\*(.*)\*", r"\\textit{\1}", s)
    return s


def append_percentage(s):
    # The percent symbol needs escaping in latex
    if rcParams["text.usetex"]:
        return s + r"$\%$"
    else:
        return s + "%"


def setup_theme(
    context="notebook",
    style="darkgrid",
    palette="deep",
    font="Helvetica",
    usetex: bool = True,
):
    try:
        extra_rc = {
            "lines.linewidth": 1,
            "lines.markeredgewidth": 1,
            "patch.edgecolor": "k",
        }
        sns.set_theme(context=context, style=style, palette=palette, rc=extra_rc)
    except (ImportError, SyntaxError):
        pass

    if usetex:
        rc("text", usetex=True)
    else:
        logger.info("Set text.usetex=%s. Font styles may be inconsistent.", usetex)
        rc("text", usetex=False)

    if font == "Helvetica":
        rc("font", **{"family": "sans-serif", "sans-serif": ["Helvetica"]})
    elif font == "Palatino":
        rc("font", **{"family": "serif", "serif": ["Palatino"]})
    elif font == "Schoolbook":
        rc("font", **{"family": "serif", "serif": ["Century Schoolbook L"]})


def asciiaxis(x, digit=1):
    if isinstance(x, int):
        x = str(x)
    elif isinstance(x, float):
        x = "{0:.{1}f}".format(x, digit)
    elif isinstance(x, np.int64):
        x = str(x)
    elif isinstance(x, np.ndarray):
        assert len(x) == 2
        x = str(x).replace("]", ")")  # upper bound not inclusive

    return x


def asciiplot(x, y, digit=1, width=50, title=None, char="="):
    """
    Print out a horizontal plot using ASCII chars.
    width is the textwidth (height) of the plot.
    """
    ax = np.array(x)
    ay = np.array(y)

    if title:
        print("[bold white]".format(title), file=sys.stderr)

    az = ay * width // ay.max()
    tx = [asciiaxis(x, digit=digit) for x in ax]
    rjust = max([len(x) for x in tx]) + 1

    for x, y, z in zip(tx, ay, az):
        x = x.rjust(rjust)
        y = y or ""
        z = "[green]{}".format(char * z)
        print("{} | {} {}".format(x, z, y), file=sys.stderr)


def print_colors(palette, outfile="Palette.png"):
    """
    print color palette (a tuple) to a PNG file for quick check
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    xmax = 20 * (len(palette) + 1)
    x1s = np.arange(0, xmax, 20)
    xintervals = [10] * len(palette)
    xx = zip(x1s, xintervals)
    ax.broken_barh(xx, (5, 10), facecolors=palette)

    ax.set_ylim(0, 20)
    ax.set_xlim(0, xmax)
    ax.set_axis_off()

    savefig(outfile)


def plot_heatmap(
    ax,
    M: np.ndarray,
    breaks: List[int],
    groups: List[Tuple[int, int, List[Tuple[int, str]], str]] = [],
    plot_breaks: bool = False,
    cmap: Optional[Union[str, Colormap]] = None,
    binsize: Optional[int] = None,
):
    """Plot heatmap illustrating the contact probabilities in Hi-C data.

    Args:
        ax (pyplot.axes): Matplotlib axis
        M (np.array): 2D numpy-array
        breaks (List[int]): Positions of chromosome starts. Can be None.
        iopts (OptionParser options): Graphical options passed in from commandline
        groups (List, optional): [(start, end, [(position, seqid)], color)]. Defaults to [].
        plot_breaks (bool): Whether to plot white breaks. Defaults to False.
        cmap (str | Colormap, optional): Colormap. Defaults to None, which uses cubehelix.
        binsize (int, optional): Resolution of the heatmap.
    """
    cmap = cmap or sns.cubehelix_palette(rot=0.5, as_cmap=True)
    ax.imshow(M, cmap=cmap, interpolation="none")
    _, xmax = ax.get_xlim()
    xlim = (0, xmax)
    if plot_breaks:
        for b in breaks[:-1]:
            ax.plot([b, b], xlim, "w-")
            ax.plot(xlim, [b, b], "w-")

    def simplify_seqid(seqid):
        seqid = seqid.replace("_", "")
        if seqid[:3].lower() == "chr":
            seqid = seqid[3:]
        return seqid.lstrip("0")

    for start, end, position_seqids, color in groups:
        # Plot a square
        ax.plot([start, start], [start, end], "-", color=color)
        ax.plot([start, end], [start, start], "-", color=color)
        ax.plot([start, end], [end, end], "-", color=color)
        ax.plot([end, end], [start, end], "-", color=color)
        for position, seqid in position_seqids:
            seqid = simplify_seqid(seqid)
            ax.text(position, end, seqid, ha="center", va="top")

    ax.set_xlim(xlim)
    ax.set_ylim((xlim[1], xlim[0]))  # Flip the y-axis so the origin is at the top
    ax.set_xticklabels(ax.get_xticks(), family="Helvetica", color="gray")
    ax.set_yticklabels(ax.get_yticks(), family="Helvetica", color="gray", rotation=90)
    ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True)
    if binsize is not None:
        formatter = ticker.FuncFormatter(
            lambda x, pos: human_readable(int(x) * binsize, pos, base=True)
        )
        ax.xaxis.set_major_formatter(formatter)
        ax.yaxis.set_major_formatter(formatter)
        title = f"Resolution = {human_size(binsize, precision=0)} per bin"
        ax.set_xlabel(title)


def discrete_rainbow(N=7, cmap=cm.Set1, usepreset=True, shuffle=False, plot=False):
    """
    Return a discrete colormap and the set of colors.

    modified from
    <http://www.scipy.org/Cookbook/Matplotlib/ColormapTransformations>

    cmap: colormap instance, eg. cm.jet.
    N: Number of colors.

    Example
    >>> x = resize(arange(100), (5,100))
    >>> djet = cmap_discretize(cm.jet, 5)
    >>> imshow(x, cmap=djet)

    See available matplotlib colormaps at:
    <http://dept.astro.lsa.umich.edu/~msshin/science/code/matplotlib_cm/>

    If N>20 the sampled colors might not be very distinctive.
    If you want to error and try anyway, set usepreset=False
    """
    import random

    from scipy import interpolate

    if usepreset:
        if 0 < N <= 5:
            cmap = cm.gist_rainbow
        elif N <= 20:
            cmap = cm.Set1
        else:
            sys.exit(discrete_rainbow.__doc__)

    cdict = cmap._segmentdata.copy()
    # N colors
    colors_i = np.linspace(0, 1.0, N)
    # N+1 indices
    indices = np.linspace(0, 1.0, N + 1)
    rgbs = []
    for key in ("red", "green", "blue"):
        # Find the N colors
        D = np.array(cdict[key])
        I = interpolate.interp1d(D[:, 0], D[:, 1])
        colors = I(colors_i)
        rgbs.append(colors)
        # Place these colors at the correct indices.
        A = np.zeros((N + 1, 3), float)
        A[:, 0] = indices
        A[1:, 1] = colors
        A[:-1, 2] = colors
        # Create a tuple for the dictionary.
        L = []
        for l in A:
            L.append(tuple(l))
        cdict[key] = tuple(L)

    palette = zip(*rgbs)

    if shuffle:
        random.shuffle(palette)

    if plot:
        print_colors(palette)

    # Return (colormap object, RGB tuples)
    return mpl.colors.LinearSegmentedColormap("colormap", cdict, 1024), palette


def get_intensity(octal):
    from math import sqrt

    r, g, b = octal[1:3], octal[3:5], octal[5:]
    r, g, b = int(r, 16), int(g, 16), int(b, 16)
    intensity = sqrt((r * r + g * g + b * b) / 3)
    return intensity


def get_shades(base_color: str, n: int) -> np.array:
    """
    Generate N shades of a base color by blending with white and black. The base
    color is in hex format, e.g., '#FF0000' for red.
    """
    base_rgb = np.array(mcolors.to_rgb(base_color))  # Convert hex to RGB (normalized)
    lighter = n // 2
    darker = n - lighter - 1

    # Generate lighter shades by blending with white
    white = np.array([1, 1, 1])  # White color as a NumPy array
    lighter_shades = [
        (white * (1 - i) + base_rgb * i) for i in np.linspace(0.2, 1, lighter + 1)
    ][:-1]

    # Generate darker shades by blending with black
    black = np.array([0, 0, 0])  # Black color as a NumPy array
    darker_shades = [
        (base_rgb * i + black * (1 - i)) for i in np.linspace(1, 0.2, darker + 1)
    ][1:]

    # Combine all shades with the base color in the middle
    return lighter_shades + [base_rgb] + darker_shades


def adjust_spines(ax, spines, outward=False, color="lightslategray"):
    # Modified from <http://matplotlib.org/examples/pylab_examples/spine_placement_demo.html>
    for loc, spine in ax.spines.items():
        if loc in spines:
            if outward:
                spine.set_position(("outward", 8))  # outward by 10 points
            spine.set_color(color)
        else:
            spine.set_color("none")  # don't draw spine

    if "left" in spines:
        ax.yaxis.set_ticks_position("left")
    else:
        ax.yaxis.set_ticks_position("right")

    if "bottom" in spines:
        ax.xaxis.set_ticks_position("bottom")
    else:
        ax.xaxis.set_ticks_position("top")

    # Change tick styles directly
    ax.tick_params(color=color)
    set_helvetica_axis(ax)


def set_ticklabels_helvetica(ax, xcast=int, ycast=int):
    xticklabels = [xcast(x) for x in ax.get_xticks()]
    yticklabels = [ycast(x) for x in ax.get_yticks()]
    ax.set_xticklabels(xticklabels, family="Helvetica")
    ax.set_yticklabels(yticklabels, family="Helvetica")


def draw_cmap(ax, cmap_text, vmin, vmax, cmap=None, reverse=False):
    # Draw a horizontal colormap at bottom-right corder of the canvas
    Y = np.outer(np.ones(10), np.arange(0, 1, 0.01))
    if reverse:
        Y = Y[::-1]
    xmin, xmax = 0.6, 0.9
    ymin, ymax = 0.02, 0.04
    ax.imshow(Y, extent=(xmin, xmax, ymin, ymax), cmap=cmap)
    ax.text(
        xmin - 0.01,
        (ymin + ymax) * 0.5,
        markup(cmap_text),
        ha="right",
        va="center",
        size=10,
    )
    vmiddle = (vmin + vmax) * 0.5
    xmiddle = (xmin + xmax) * 0.5
    for x, v in zip((xmin, xmiddle, xmax), (vmin, vmiddle, vmax)):
        ax.text(x, ymin - 0.005, "%.1f" % v, ha="center", va="top", size=10)


def write_messages(ax, messages: List[str], ypad: float = 0.04):
    """
    Write text on canvas, usually on the top right corner.
    """
    tc = "gray"
    axt = ax.transAxes
    yy = 0.95
    for msg in messages:
        ax.text(0.95, yy, markup(msg), color=tc, transform=axt, ha="right")
        yy -= ypad


def quickplot_ax(
    ax,
    data,
    xmin,
    xmax,
    xlabel,
    title=None,
    ylabel="Counts",
    counts=True,
    percentage=True,
    highlight=None,
):
    # TODO: redundant with quickplot(), need to be refactored.
    if percentage:
        total_length = sum(data.values())
        data = dict((k, v * 100.0 / total_length) for (k, v) in data.items())

    left, height = zip(*sorted(data.items()))
    pad = max(height) * 0.01
    c1, c2 = "darkslategray", "tomato"
    if counts:
        for l, h in zip(left, height):
            if xmax and l > xmax:
                break
            tag = str(int(h))
            rotation = 90
            if percentage:
                tag = append_percentage(tag) if int(tag) > 0 else ""
                rotation = 0
            color = c1
            if highlight is not None and l in highlight:
                color = c2
            ax.text(
                l,
                h + pad,
                tag,
                color=color,
                size=8,
                ha="center",
                va="bottom",
                rotation=rotation,
            )
    if xmax is None:
        xmax = max(left)

    ax.bar(left, height, align="center", fc=c1)
    if highlight:
        for h in highlight:
            ax.bar([h], [data[h]], align="center", ec=c2, fc=c2)

    ax.set_xlabel(markup(xlabel))
    if ylabel:
        ax.set_ylabel(markup(ylabel))
    if title:
        ax.set_title(markup(title))
    ax.set_xlim((xmin - 0.5, xmax + 0.5))
    if percentage:
        ax.set_ylim(0, 100)


def quickplot(
    data,
    xmin,
    xmax,
    xlabel,
    title,
    ylabel="Counts",
    figname="plot.pdf",
    counts=True,
    print_stats=True,
):
    """
    Simple plotting function - given a dictionary of data, produce a bar plot
    with the counts shown on the plot.
    """
    plt.figure(1, (6, 6))
    left, height = zip(*sorted(data.items()))
    pad = max(height) * 0.01
    if counts:
        for l, h in zip(left, height):
            if xmax and l > xmax:
                break
            plt.text(
                l,
                h + pad,
                str(h),
                color="darkslategray",
                size=8,
                ha="center",
                va="bottom",
                rotation=90,
            )
    if xmax is None:
        xmax = max(left)

    plt.bar(left, height, align="center")
    plt.xlabel(markup(xlabel))
    plt.ylabel(markup(ylabel))
    plt.title(markup(title))
    plt.xlim((xmin - 0.5, xmax + 0.5))

    # Basic statistics
    messages = []
    counts_over_xmax = sum([v for k, v in data.items() if k > xmax])
    if counts_over_xmax:
        messages += ["Counts over xmax({0}): {1}".format(xmax, counts_over_xmax)]
    kk = []
    for k, v in data.items():
        kk += [k] * v
    messages += ["Total: {0}".format(np.sum(height))]
    messages += ["Maximum: {0}".format(np.max(kk))]
    messages += ["Minimum: {0}".format(np.min(kk))]
    messages += ["Average: {0:.2f}".format(np.mean(kk))]
    messages += ["Median: {0}".format(np.median(kk))]
    ax = plt.gca()
    if print_stats:
        write_messages(ax, messages)

    set_human_axis(ax)
    set_ticklabels_helvetica(ax)
    savefig(figname)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
SynFind analyses and visualization.
"""
from collections import defaultdict
from copy import deepcopy
from itertools import groupby
import os.path as op
import sys

from ..apps.base import ActionDispatcher, OptionParser, logger, mkdir, symlink
from ..apps.grid import MakeManager
from ..formats.base import get_number, must_open
from ..formats.bed import Bed
from ..formats.blast import BlastLine
from ..formats.gff import Gff, load
from ..graphics.base import (
    FancyArrow,
    latex,
    markup,
    normalize_axes,
    panel_labels,
    plt,
    savefig,
)
from ..graphics.glyph import CartoonRegion, RoundRect
from ..utils.cbook import SummaryStats, gene_name, percentage
from ..utils.grouper import Grouper


def main():

    actions = (
        ("cartoon", "generate cartoon illustration of SynFind"),
        ("ecoli", "gene presence absence analysis in ecoli"),
        ("grass", "validate SynFind pan-grass set against James"),
        ("coge", "prepare coge datasets"),
        # For benchmarking
        ("synfind", "prepare input for SynFind"),
        ("iadhore", "prepare input for iADHoRe"),
        ("mcscanx", "prepare input for MCScanX"),
        ("cyntenator", "prepare input for Cyntenator"),
        ("athalianatruth", "prepare truth pairs for At alpha/beta/gamma"),
        ("yeasttruth", "prepare truth pairs for 14 yeasts"),
        ("grasstruth", "prepare truth pairs for 4 grasses"),
        ("benchmark", "compare SynFind, MCScanX, iADHoRe and OrthoFinder"),
        ("venn", "display benchmark results as Venn diagram"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def grasstruth(args):
    """
    %prog grasstruth james-pan-grass.txt

    Prepare truth pairs for 4 grasses.
    """
    p = OptionParser(grasstruth.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (james,) = args
    fp = open(james)
    pairs = set()
    for row in fp:
        atoms = row.split()
        genes = []
        idx = {}
        for i, a in enumerate(atoms):
            aa = a.split("||")
            for ma in aa:
                idx[ma] = i
            genes.extend(aa)
        genes = [x for x in genes if ":" not in x]
        Os = [x for x in genes if x.startswith("Os")]
        for o in Os:
            for g in genes:
                if idx[o] == idx[g]:
                    continue
                pairs.add(tuple(sorted((o, g))))

    for a, b in sorted(pairs):
        print("\t".join((a, b)))


def synfind(args):
    """
    %prog synfind all.last *.bed

    Prepare input for SynFind.
    """
    p = OptionParser(synfind.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    lastfile = args[0]
    bedfiles = args[1:]
    fp = open(lastfile)
    filteredlast = lastfile + ".filtered"
    fw = open(filteredlast, "w")
    for row in fp:
        b = BlastLine(row)
        if b.query == b.subject:
            continue
        print(b, file=fw)
    fw.close()
    logger.debug("Filtered LAST file written to `{0}`".format(filteredlast))

    allbed = "all.bed"
    fw = open(allbed, "w")
    for i, bedfile in enumerate(bedfiles):
        prefix = chr(ord("A") + i)
        bed = Bed(bedfile)
        for b in bed:
            b.seqid = prefix + b.seqid
            print(b, file=fw)
    fw.close()
    logger.debug("Bed file written to `{0}`".format(allbed))


def yeasttruth(args):
    """
    %prog yeasttruth Pillars.tab *.gff

    Prepare pairs data for 14 yeasts.
    """
    p = OptionParser(yeasttruth.__doc__)
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    pillars = args[0]
    gffiles = args[1:]
    aliases = {}
    pivot = {}
    for gffile in gffiles:
        is_pivot = op.basename(gffile).startswith("Saccharomyces_cerevisiae")
        gff = Gff(gffile)
        for g in gff:
            if g.type != "gene":
                continue
            for a in g.attributes["Alias"]:
                aliases[a] = g.accn
                if is_pivot:
                    pivot[a] = g.accn
    logger.debug("Aliases imported: {0}".format(len(aliases)))
    logger.debug("Pivot imported: {0}".format(len(pivot)))
    fw = open("yeast.aliases", "w")
    for k, v in sorted(aliases.items()):
        print("\t".join((k, v)), file=fw)
    fw.close()

    fp = open(pillars)
    pairs = set()
    fw = must_open(opts.outfile, "w")
    for row in fp:
        atoms = [x for x in row.split() if x != "---"]
        pps = [pivot[x] for x in atoms if x in pivot]
        atoms = [aliases[x] for x in atoms if x in aliases]
        for p in pps:
            for a in atoms:
                if p == a:
                    continue
                pairs.add(tuple(sorted((p, a))))

    for a, b in sorted(pairs):
        print("\t".join((a, b)), file=fw)
    fw.close()


def venn(args):
    """
    %prog venn *.benchmark

    Display benchmark results as Venn diagram.
    """
    from matplotlib_venn import venn2

    p = OptionParser(venn.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="9x9")

    if len(args) < 1:
        sys.exit(not p.print_help())

    bcs = args
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    pad = 0.02
    ystart = 1
    ywidth = 1.0 / len(bcs)
    tags = ("Bowers", "YGOB", "Schnable")
    for bc, tag in zip(bcs, tags):
        fp = open(bc)
        data = []
        for row in fp:
            prog, pcounts, tcounts, shared = row.split()
            pcounts = int(pcounts)
            tcounts = int(tcounts)
            shared = int(shared)
            data.append((prog, pcounts, tcounts, shared))
        xstart = 0
        xwidth = 1.0 / len(data)
        for prog, pcounts, tcounts, shared in data:
            a, b, c = pcounts - shared, tcounts - shared, shared
            ax = fig.add_axes(
                [
                    xstart + pad,
                    ystart - ywidth + pad,
                    xwidth - 2 * pad,
                    ywidth - 2 * pad,
                ]
            )
            venn2(subsets=(a, b, c), set_labels=(prog, tag), ax=ax)
            message = "Sn={0} Pu={1}".format(
                percentage(shared, tcounts, precision=0, mode=-1),
                percentage(shared, pcounts, precision=0, mode=-1),
            )
            print(message, file=sys.stderr)
            ax.text(
                0.5,
                0.92,
                latex(message),
                ha="center",
                va="center",
                transform=ax.transAxes,
                color="b",
            )
            ax.set_axis_off()
            xstart += xwidth
        ystart -= ywidth

    panel_labels(
        root,
        ((0.04, 0.96, "A"), (0.04, 0.96 - ywidth, "B"), (0.04, 0.96 - 2 * ywidth, "C")),
    )
    panel_labels(
        root,
        (
            (0.5, 0.98, "A. thaliana duplicates"),
            (0.5, 0.98 - ywidth, "14 Yeast genomes"),
            (0.5, 0.98 - 2 * ywidth, "4 Grass genomes"),
        ),
    )
    normalize_axes(root)
    savefig("venn.pdf", dpi=opts.dpi)


def coge(args):
    """
    %prog coge *.gff

    Prepare coge datasets.
    """
    p = OptionParser(coge.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    gffs = args
    for gff in gffs:
        atoms = op.basename(gff).split(".")
        gid = atoms[-2]
        assert gid.startswith("gid")
        gid = get_number(gid)
        genomefasta = "genome_{0}.faa.fasta".format(gid)
        species = "_".join(atoms[0].split("_")[:2])
        cdsfasta = species + ".cds.fasta"
        load(
            [
                gff,
                genomefasta,
                "--id_attribute=Parent",
                "--outfile={0}".format(cdsfasta),
            ]
        )


def calc_sensitivity_specificity(a, truth, tag, fw):
    common = a & truth
    sensitivity = len(common) * 100.0 / len(truth)
    specificity = len(common) * 100.0 / len(a)
    logger.debug("{0}: {1} pairs".format(tag, len(a)))
    logger.debug(
        "{0}: Sensitivity={1:.1f}% Purity={2:.1f}%".format(
            tag, sensitivity, specificity
        )
    )
    print(tag, len(a), len(truth), len(common), file=fw)


def write_pairs(pairs, pairsfile):
    fz = open(pairsfile, "w")
    for a, b in pairs:
        print("\t".join((a, b)), file=fz)
    fz.close()


def benchmark(args):
    """
    %prog benchmark at bedfile

    Compare SynFind, MCScanx, iADHoRe and OrthoFinder against the truth.
    """
    p = OptionParser(benchmark.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    pf, bedfile = args
    truth = pf + ".truth"
    synfind = pf + ".synfind"
    mcscanx = pf + ".mcscanx"
    iadhore = pf + ".iadhore"
    orthofinder = pf + ".orthofinder"
    pivots = set([x.accn for x in Bed(bedfile)])

    fp = open(truth)
    truth = set()
    for row in fp:
        a, b = row.strip().split("\t")[:2]
        pivots.add(a)
        truth.add(tuple(sorted((a, b))))
    logger.debug("Truth: {0} pairs".format(len(truth)))

    fp = open(synfind)
    benchmarkfile = pf + ".benchmark"
    fw = must_open(benchmarkfile, "w")
    synfind = set()
    for row in fp:
        atoms = row.strip().split("\t")
        query, hit, tag = atoms[:3]
        if tag != "S":
            continue
        synfind.add(tuple(sorted((query, hit))))
    calc_sensitivity_specificity(synfind, truth, "SynFind", fw)

    fp = open(mcscanx)
    mcscanx = set()
    for row in fp:
        if row[0] == "#":
            continue
        atoms = row.strip().split(":")[1].split()
        query, hit = atoms[:2]
        mcscanx.add(tuple(sorted((query, hit))))
    calc_sensitivity_specificity(mcscanx, truth, "MCScanX", fw)

    fp = open(iadhore)
    iadhore = set()
    next(fp)
    for row in fp:
        atoms = row.strip().split("\t")
        query, hit = atoms[3:5]
        iadhore.add(tuple(sorted((query, hit))))
    calc_sensitivity_specificity(iadhore, truth, "iADHoRe", fw)

    fp = open(orthofinder)
    orthofinder = set()
    next(fp)
    for row in fp:
        row = row.replace('"', "")
        atoms = row.replace(",", " ").split()
        genes = [x.strip() for x in atoms if not x.startswith("OG")]
        genes = [gene_name(x) for x in genes]
        pps = [x for x in genes if x in pivots]
        for p in pps:
            for g in genes:
                if p == g:
                    continue
                orthofinder.add(tuple(sorted((p, g))))
    # write_pairs(orthofinder, "orthofinder.pairs")
    calc_sensitivity_specificity(orthofinder, truth, "OrthoFinder", fw)
    fw.close()


def write_lst(bedfile):
    pf = op.basename(bedfile).split(".")[0]
    mkdir(pf)
    bed = Bed(bedfile)
    stanza = []
    for seqid, bs in bed.sub_beds():
        fname = op.join(pf, "{0}.lst".format(seqid))
        fw = open(fname, "w")
        for b in bs:
            print("{0}{1}".format(b.accn.replace(" ", ""), b.strand), file=fw)
        stanza.append((seqid, fname))
        fw.close()
    return pf, stanza


def write_txt(bedfile):
    pf = op.basename(bedfile).split(".")[0][:20]
    txtfile = pf + ".txt"
    fw = open(txtfile, "w")
    print("#genome", file=fw)
    bed = Bed(bedfile)
    for b in bed:
        print(
            " ".join(str(x) for x in (b.accn, b.seqid, b.start, b.end, b.strand)),
            file=fw,
        )
    fw.close()
    return txtfile


def cyntenator(args):
    """
    %prog cyntenator athaliana.athaliana.last athaliana.bed

    Prepare input for Cyntenator.
    """
    p = OptionParser(cyntenator.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    lastfile = args[0]
    fp = open(lastfile)
    filteredlastfile = lastfile + ".blast"
    fw = open(filteredlastfile, "w")
    for row in fp:
        b = BlastLine(row)
        if b.query == b.subject:
            continue
        print("\t".join((b.query, b.subject, str(b.score))), file=fw)
    fw.close()

    bedfiles = args[1:]
    fp = open(lastfile)
    b = BlastLine(next(fp))
    subject = b.subject
    txtfiles = []
    for bedfile in bedfiles:
        order = Bed(bedfile).order
        if subject in order:
            db = op.basename(bedfile).split(".")[0][:20]
            logger.debug("Found db: {0}".format(db))
        txtfile = write_txt(bedfile)
        txtfiles.append(txtfile)

    db += ".txt"
    mm = MakeManager()
    for txtfile in txtfiles:
        outfile = txtfile + ".alignment"
        cmd = 'cyntenator -t "({0} {1})" -h blast {2} > {3}'.format(
            txtfile, db, filteredlastfile, outfile
        )
        mm.add((txtfile, db, filteredlastfile), outfile, cmd)
    mm.write()


def iadhore(args):
    """
    %prog iadhore athaliana.athaliana.last athaliana.bed

    Wrap around iADHoRe.
    """
    p = OptionParser(iadhore.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    lastfile = args[0]
    bedfiles = args[1:]
    blast_table = "blast_table.txt"
    fp = open(lastfile)
    seen = set()
    for row in fp:
        c = BlastLine(row)
        a, b = c.query, c.subject
        a, b = gene_name(a), gene_name(b)
        if a > b:
            a, b = b, a
        seen.add((a, b))

    fw = open(blast_table, "w")
    for a, b in seen:
        print("\t".join((a, b)), file=fw)
    fw.close()
    logger.debug("A total of {0} pairs written to `{1}`".format(len(seen), blast_table))

    fw = open("config.txt", "w")
    for bedfile in bedfiles:
        pf, stanza = write_lst(bedfile)
        print("genome={0}".format(pf), file=fw)
        for seqid, fname in stanza:
            print(" ".join((seqid, fname)), file=fw)
        print(file=fw)

    print("blast_table={0}".format(blast_table), file=fw)
    print("cluster_type=colinear", file=fw)
    print("tandem_gap=10", file=fw)
    print("prob_cutoff=0.001", file=fw)
    print("gap_size=20", file=fw)
    print("cluster_gap=20", file=fw)
    print("q_value=0.9", file=fw)
    print("anchor_points=4", file=fw)
    print("alignment_method=gg2", file=fw)
    print("max_gaps_in_alignment=20", file=fw)
    print("output_path=i-adhore_out", file=fw)
    print("number_of_threads=4", file=fw)
    fw.close()


def extract_groups(g, pairs, txtfile):
    register = defaultdict(list)
    fp = open(txtfile)
    next(fp)
    for row in fp:
        if row[0] != ">":
            continue
        track, atg, myname, pairname = row.split()
        pairname = pairname.rstrip("ab").upper()
        register[pairname].append(atg.upper())

    for pairname, genes in register.items():
        tag = pairname[0]
        tag = {"A": "alpha", "B": "beta", "C": "gamma", "S": "others"}[tag]
        pairs.add(tuple(sorted(genes) + [tag]))
        g.join(*genes)


def athalianatruth(args):
    """
    %prog athalianatruth J_a.txt J_bc.txt

    Prepare pairs data for At alpha/beta/gamma.
    """
    p = OptionParser(athalianatruth.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    atxt, bctxt = args
    g = Grouper()
    pairs = set()
    for txt in (atxt, bctxt):
        extract_groups(g, pairs, txt)

    fw = open("pairs", "w")
    for pair in sorted(pairs):
        print("\t".join(pair), file=fw)
    fw.close()

    fw = open("groups", "w")
    for group in list(g):
        print(",".join(group), file=fw)
    fw.close()


def make_gff(bed, prefix, fw):
    bed = Bed(bed)
    nfeats = 0
    for b in bed:
        seqid = prefix + b.seqid
        print("\t".join(str(x) for x in (seqid, b.accn, b.start, b.end)), file=fw)
        nfeats += 1
    logger.debug("A total of {0} features converted to `{1}`".format(nfeats, fw.name))


def mcscanx(args):
    """
    %prog mcscanx athaliana.athaliana.last athaliana.bed

    Wrap around MCScanX.
    """
    p = OptionParser(mcscanx.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    blastfile = args[0]
    bedfiles = args[1:]
    prefix = "_".join(op.basename(x)[:2] for x in bedfiles)
    symlink(blastfile, prefix + ".blast")
    allbedfile = prefix + ".gff"
    fw = open(allbedfile, "w")
    for i, bedfile in enumerate(bedfiles):
        prefix = chr(ord("A") + i)
        make_gff(bedfile, prefix, fw)
    fw.close()


def grass(args):
    """
    %prog grass coge_master_table.txt james.txt

    Validate SynFind pan-grass set against James. This set can be generated:

    https://genomevolution.org/r/fhak
    """
    p = OptionParser(grass.__doc__)
    p.set_verbose()
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    master, james = args

    fp = open(master)
    next(fp)
    master_store = defaultdict(set)
    for row in fp:
        atoms = row.split()
        s = set()
        for x in atoms[1:6]:
            m = x.split(",")
            s |= set(m)
        if "-" in s:
            s.remove("-")

        a = atoms[1]
        master_store[a] |= set(s)

    fp = open(james)
    next(fp)
    james_store = {}
    tandems = set()
    for row in fp:
        atoms = row.split()
        s = set()
        Os = set()
        for x in atoms[:-1]:
            m = x.split("||")
            if m[0].startswith("Os"):
                Os |= set(m)
            if m[0].startswith("http"):
                continue
            if m[0].startswith("chr"):
                m = ["proxy"]
            if "||" in x:
                tandems |= set(m)
            s |= set(m)

        for x in Os:
            james_store[x] = s

    jaccards = []
    corr_jaccards = []
    perfect_matches = 0
    corr_perfect_matches = 0
    for k, v in james_store.items():
        if k not in master_store:
            continue
        m = master_store[k]
        jaccard = len(v & m) * 100 / len(v | m)
        jaccards.append(jaccard)
        diff = (v ^ m) - tandems
        corr_jaccard = 100 - len(diff) * 100 / len(v | m)
        corr_jaccards.append(corr_jaccard)
        if opts.verbose:
            print(k)
            print(v)
            print(m)
            print(diff)
            print(jaccard)
        if jaccard > 99:
            perfect_matches += 1
        if corr_jaccard > 99:
            corr_perfect_matches += 1

    logger.debug("Perfect matches: {0}".format(perfect_matches))
    logger.debug("Perfect matches (corrected): {0}".format(corr_perfect_matches))
    print("Jaccards:", SummaryStats(jaccards))
    print("Corrected Jaccards:", SummaryStats(corr_jaccards))


def ecoli(args):
    """
    %prog ecoli coge_master_table.txt query.bed

    Perform gene presence / absence analysis in Ecoli master spreadsheet. Ecoli
    spresheets can be downloaded below:

    Ecoli K12 MG1655 (K) as query
    Regenerate this analysis: https://genomevolution.org/r/fggo

    Ecoli O157:H7 EDL933 (O) as query
    Regenerate this analysis: https://genomevolution.org/r/fgt7

    Shigella flexneri 2a 301 (S) as query
    Regenerate this analysis: https://genomevolution.org/r/fgte

    Perform a similar analysis as in:
    Jin et al. (2002) Genome sequence of Shigella flexneri 2a: insights
    into pathogenicity through comparison with genomes of Escherichia
    coli K12 and O157. Nucleic Acid Research.
    """
    p = OptionParser(ecoli.__doc__)
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    master, querybed = args
    fp = open(master)
    header = next(fp)
    assert header[0] == "#"
    qorg = header.strip().split("\t")[1]
    qorg = qorg.split(":")[-1].strip()

    store = {}
    MISSING = ("proxy", "-")
    for row in fp:
        a, b, c = row.strip().split("\t")[1:4]
        store[a] = b in MISSING and c in MISSING

    bed = Bed(querybed)
    tags = []
    for i, b in enumerate(bed):
        accn = b.accn
        if accn not in store:
            logger.warning("missing {0}".format(accn))
            continue
        tags.append((store[accn], accn))

    large = 4  # large segments
    II = []
    II_large = []
    for missing, aa in groupby(tags, key=lambda x: x[0]):
        aa = list(aa)
        if not missing:
            continue
        glist = list(a for missing, a in aa)
        II.append(glist)
        size = len(glist)
        if size >= large:
            II_large.append(glist)

    fw = must_open(opts.outfile, "w")
    for a, t in zip((II, II_large), ("", ">=4 ")):
        nmissing = sum(len(x) for x in a)
        logger.debug(
            "A total of {0} {1}-specific {2}islands found with {3} genes.".format(
                len(a), qorg, t, nmissing
            )
        )

    for x in II:
        print(len(x), ",".join(x), file=fw)


def plot_diagram(ax, x, y, A, B, tag, label):
    ax.text(x, y + 0.14, "{0}: {1}".format(tag, label), ha="center")
    strip = tag != "G"
    A.draw(ax, x, y + 0.06, gene_len=0.02, strip=strip)
    B.draw(ax, x, y, gene_len=0.02, strip=strip)


def cartoon(args):
    """
    %prog synteny.py

    Generate cartoon illustration of SynFind.
    """
    p = OptionParser(cartoon.__doc__)
    opts, args, iopts = p.set_image_options(args, figsize="10x7")

    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])

    # Panel A
    A = CartoonRegion(41)
    A.draw(root, 0.35, 0.85, strip=False, color=False)
    x1, x2 = A.x1, A.x2
    lsg = "lightslategray"
    pad = 0.01
    xc, yc = 0.35, 0.88
    arrowlen = x2 - xc - pad
    arrowprops = dict(
        length_includes_head=True,
        width=0.01,
        fc=lsg,
        lw=0,
        head_length=arrowlen * 0.15,
        head_width=0.03,
    )
    p = FancyArrow(xc - pad, yc, -arrowlen, 0, shape="left", **arrowprops)
    root.add_patch(p)
    p = FancyArrow(xc + pad, yc, arrowlen, 0, shape="right", **arrowprops)
    root.add_patch(p)

    yt = yc + 4 * pad
    root.text((x1 + xc) / 2, yt, "20 genes upstream", ha="center")
    root.text((x2 + xc) / 2, yt, "20 genes downstream", ha="center")
    root.plot((xc,), (yc,), "o", mfc="w", mec=lsg, mew=2, lw=2, color=lsg)
    root.text(xc, yt, "Query gene", ha="center")

    # Panel B
    A.draw(root, 0.35, 0.7, strip=False)

    RoundRect(root, (0.07, 0.49), 0.56, 0.14, fc="y", alpha=0.2)
    a = deepcopy(A)
    a.evolve(mode="S", target=10)
    a.draw(root, 0.35, 0.6)
    b = deepcopy(A)
    b.evolve(mode="F", target=8)
    b.draw(root, 0.35, 0.56)
    c = deepcopy(A)
    c.evolve(mode="G", target=6)
    c.draw(root, 0.35, 0.52)

    for x in (a, b, c):
        root.text(0.64, x.y, "Score={0}".format(x.nonwhites), va="center")

    # Panel C
    A.truncate_between_flankers()
    a.truncate_between_flankers()
    b.truncate_between_flankers()
    c.truncate_between_flankers(target=6)

    plot_diagram(root, 0.14, 0.2, A, a, "S", "syntenic")
    plot_diagram(root, 0.37, 0.2, A, b, "F", "missing, with both flankers")
    plot_diagram(root, 0.6, 0.2, A, c, "G", "missing, with one flanker")

    labels = ((0.04, 0.95, "A"), (0.04, 0.75, "B"), (0.04, 0.4, "C"))
    panel_labels(root, labels)

    # Descriptions
    xt = 0.85
    desc = (
        "Extract neighborhood",
        "of *window* size",
        "Count gene pairs within *window*",
        "Find regions above *score* cutoff",
        "Identify flankers",
        "Annotate syntelog class",
    )
    for yt, t in zip((0.88, 0.84, 0.64, 0.6, 0.3, 0.26), desc):
        root.text(xt, yt, markup(t), ha="center", va="center")

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    pf = "cartoon"
    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Create chromosome landscape plots that are similar to the ones used in soybean
and sorghum paper.
"""


from collections import Counter, OrderedDict, defaultdict
import os.path as op
import sys
from typing import Dict, List, Optional, Tuple

import numpy as np
import seaborn as sns

from ..algorithms.matrix import moving_sum
from ..apps.base import ActionDispatcher, OptionParser, logger
from ..formats.base import BaseFile, DictFile, LineFile, must_open
from ..formats.bed import Bed, bins, get_nbins
from ..formats.sizes import Sizes
from ..utils.cbook import autoscale, human_size, percentage
from .base import (
    CirclePolygon,
    Colormap,
    Extent,
    Rectangle,
    adjust_extent,
    adjust_spines,
    human_readable_base,
    latex,
    markup,
    normalize_axes,
    plt,
    savefig,
    set_human_axis,
    ticker,
)
from .chromosome import HorizontalChromosome

# Colors picked from Schmutz soybean genome paper using ColorPic
palette = ["#ACABD5", "#DBF0F5", "#3EA77A", "#FBF5AB", "#C162A6"] + list("rgbymck")
gray = "#CCCCCB"
Registration = {
    "Gypsy": "LTR-RT/Gypsy",
    "Copia": "LTR-RT/Copia",
    "hAT": "DNA-TE/hAT",
    "Helitron": "DNA-TE/Helitron",
    "Tourist": "DNA-TE/Tourist",
    "Introns": "Genes (introns)",
    "Exons": "Genes (exons)",
}

# Consider a depth of 5 as minimum covered depth
MIN_COVERED_DEPTH = 5


class BinLine:
    def __init__(self, row):
        args = row.split()
        self.chr = args[0]
        self.len = float(args[1])
        self.binlen = int(args[2])

    def __str__(self):
        return "\t".join(str(x) for x in (self.chr, self.len, self.binlen))

    def subtract(self, o):
        self.binlen -= o.len


class BinFile(LineFile):
    def __init__(self, filename):
        super().__init__(filename)
        self.mapping = defaultdict(list)

        fp = open(filename, encoding="utf-8")
        for row in fp:
            b = BinLine(row)
            self.append(b)
            chr, len, binlen = b.chr, b.len, b.binlen
            self.mapping[chr].append((len, binlen))
        fp.close()


class ChrInfoLine:
    def __init__(self, row, delimiter=","):
        args = [x.strip() for x in row.split(delimiter)]
        self.name = args[0]
        self.color = args[1]
        if len(args) > 2:
            self.new_name = args[2]
        else:
            self.new_name = self.name


class ChrInfoFile(BaseFile, OrderedDict):
    def __init__(self, filename, delimiter=","):
        super().__init__(filename)
        with open(filename, encoding="utf-8") as fp:
            for row in fp:
                if row[0] == "#":
                    continue
                line = ChrInfoLine(row, delimiter=delimiter)
                self[line.name] = line


class TitleInfoLine:
    def __init__(self, row, delimiter=","):
        args = [x.strip() for x in row.split(delimiter)]
        self.name = args[0]
        self.title = args[1]
        self.subtitle = None
        if len(args) > 2:
            self.subtitle = args[2]


class TitleInfoFile(BaseFile, OrderedDict):
    def __init__(self, filename, delimiter=","):
        super().__init__(filename)
        with open(filename, encoding="utf-8") as fp:
            for row in fp:
                if row[0] == "#":
                    continue
                line = TitleInfoLine(row, delimiter=delimiter)
                self[line.name] = line


def main():

    actions = (
        ("composite", "combine line plots, feature bars and alt-bars"),
        ("depth", "show per chromosome depth plot across genome"),
        ("heatmap", "similar to stack but adding heatmap"),
        ("mosdepth", "plot depth vs. coverage per chromosome"),
        ("multilineplot", "combine multiple line plots in one vertical stack"),
        ("stack", "create landscape plot with genic/te composition"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def parse_distfile(filename):
    """Parse mosdepth dist.txt file. The file has contents like:

    #chr    start   end     depth   (header added here for clarity)
    chr01A  0       50000   31.00
    chr01A  50000   100000  36.00
    chr01A  100000  150000  280.00
    chr01A  150000  200000  190.00

    Args:
        filename (str): Path to the file.
    """
    dists = defaultdict(Counter)
    with must_open(filename) as fp:
        for row in fp:
            chromosome, _, _, depth = row.split()
            depth = int(float(depth))
            dists[chromosome][depth] += 1
    logger.debug("Loaded %d seqids", len(dists))
    return dists


def parse_groupsfile(filename):
    """Parse groupsfile, which contains the tracks to be plotted
    in the vertically stacked mosdepth plot.

    chr01A,chr01B g,m
    chr02A,chr02B g,m
    chr03A,chr03B g,m

    Args:
        filename (str): Path to the groups file.
    """
    groups = []
    with open(filename, encoding="utf-8") as fp:
        for row in fp:
            chrs, colors = row.split()
            groups.append((chrs.split(","), colors.split(",")))
    logger.debug("Loaded %d groups", len(groups))
    return groups


def cumarray_to_array(ar):
    """Convert cumulative array to normal array.

    Args:
        ar (List): List of numbers
    """
    ans = []
    for i, x in enumerate(ar):
        ans.append(x if i == 0 else (ar[i] - ar[i - 1]))
    return ans


def mosdepth(args):
    """
    %prog mosdepth mosdepth.global.dist.txt groups

    Plot depth vs. coverage per chromosome. Inspired by mosdepth plot. See also:
    https://github.com/brentp/mosdepth
    """
    sns.set_style("darkgrid")

    p = OptionParser(mosdepth.__doc__)
    p.add_argument("--maxdepth", default=100, type=int, help="Maximum depth to plot")
    p.add_argument(
        "--logscale", default=False, action="store_true", help="Use log-scale on depth"
    )
    opts, args, iopts = p.set_image_options(args, style="dark", figsize="6x8")

    if len(args) != 2:
        sys.exit(p.print_help())

    # Read in datasets
    distfile, groupsfile = args
    dists = parse_distfile(distfile)
    groups = parse_groupsfile(groupsfile)
    logscale = opts.logscale

    # Construct a composite figure with N tracks indicated in the groups
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes((0, 0, 1, 1))

    rows = len(groups)
    ypad = 0.05
    yinterval = (1 - 2 * ypad) / (rows + 1)
    yy = 1 - ypad

    for group_idx, (chrs, colors) in enumerate(groups):
        yy -= yinterval
        ax = fig.add_axes((0.15, yy, 0.7, yinterval * 0.85))
        for c, color in zip(chrs, colors):
            cdata = dists[c].items()
            logger.debug("Importing %d records for %s", len(cdata), c)
            cx, cy = zip(*sorted(cdata))
            ax.plot(cx, cy, "-", color=color)
        if logscale:
            ax.set_xscale("log", basex=2)
        ax.set_xlim(1 if logscale else 0, opts.maxdepth)
        ax.get_yaxis().set_visible(False)
        if group_idx != rows - 1:
            ax.get_xaxis().set_visible(False)

        # Add legend to the right of the canvas
        label_pad = 0.02
        label_yy = yy + yinterval
        for c, color in zip(chrs, colors):
            label_yy -= label_pad
            root.text(0.92, label_yy, c, color=color, ha="center", va="center")

    root.text(
        0.1,
        0.5,
        "Proportion of bases at coverage",
        rotation=90,
        color="darkslategray",
        ha="center",
        va="center",
    )
    root.text(0.5, 0.05, "Coverage", color="darkslategray", ha="center", va="center")
    normalize_axes(root)
    adjust_spines(ax, ["bottom"], outward=True)

    pf = "mosdepth"
    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def draw_depth(
    root,
    ax,
    bed: Bed,
    chrinfo: dict = {},
    defaultcolor: str = "k",
    sepcolor: str = "w",
    maxdepth: int = 100,
    logscale: bool = False,
    title: Optional[str] = None,
    subtitle: Optional[str] = None,
    median_line: bool = True,
    draw_seqids: bool = True,
    calculate_coverage: bool = False,
    roi: Optional[List[Tuple[str, int]]] = None,
    left_margin: float = 0.1,
    right_margin: float = 0.15,
):
    """Draw depth plot on the given axes, using data from bed

    Args:
        root (matplotlib.Axes): Canvas axes
        ax (matplotlib.Axes): Axes to plot data on
        bed (Bed): Bed data from mosdepth
        chrinfo (ChrInfoFile): seqid => color, new name
        defaultcolor (str): matplotlib-compatible color for data points
        sepcolor (str): matplotlib-compatible color for chromosome breaks
        maxdepth (int): Upper limit of the y-axis (depth)
        title (str): Title of the figure, to the right of the axis
        subtitle (str): Subtitle of the figure, just below title
    """
    if chrinfo is None:
        chrinfo = {}
    sizes = bed.max_bp_in_chr
    seqids = chrinfo.keys() if chrinfo else sizes.keys()
    starts = {}
    ends = {}
    label_positions = []
    start = 0
    end = 0
    for seqid in seqids:
        if seqid not in sizes:
            continue
        starts[seqid] = start
        end = start + sizes[seqid]
        ends[seqid] = end
        label_positions.append((seqid, (start + end) / 2))
        start = end
    xsize = end

    # Extract plotting data
    data = []
    data_by_seqid = defaultdict(list)
    total_bp = 0
    covered_bp = 0
    for b in bed:
        seqid = b.seqid
        if seqid not in starts:
            continue
        # chr01A  2000000 3000000 113.00
        x = starts[seqid] + (b.start + b.end) / 2
        y = float(b.accn)
        c = chrinfo[seqid].color if seqid in chrinfo else "k"
        data.append((x, y, c))
        data_by_seqid[seqid].append(y)
        if y >= MIN_COVERED_DEPTH:
            covered_bp += b.end - b.start
        total_bp += b.end - b.start
    logger.debug("cov: %s", percentage(covered_bp, total_bp, precision=0))

    x, y, c = zip(*data)
    ax.scatter(
        x,
        y,
        c=c,
        edgecolors="none",
        s=8,
        lw=0,
    )
    logger.debug("Obtained %d data points with depth data", len(data))

    # Per seqid median
    medians = {}
    for seqid, values in data_by_seqid.items():
        c = chrinfo[seqid].color if seqid in chrinfo else defaultcolor
        seqid_start = starts[seqid]
        seqid_end = ends[seqid]
        seqid_median = np.median(values)
        medians[seqid] = seqid_median
        if median_line:
            ax.plot(
                (seqid_start, seqid_end),
                (seqid_median, seqid_median),
                "-",
                lw=4,
                color=c,
                alpha=0.5,
            )

    # Vertical lines for all the breaks
    for pos in starts.values():
        ax.plot((pos, pos), (0, maxdepth), "-", lw=1, color=sepcolor)

    # Beautify the numeric axis
    for tick in ax.get_xticklines() + ax.get_yticklines():
        tick.set_visible(False)

    median_depth_y = 0.88
    chr_label_y = 0.08
    width = 1 - left_margin - right_margin
    rotation = 20 if len(label_positions) > 10 else 0
    for seqid, position in label_positions:
        xpos = left_margin + position * width / xsize
        c = chrinfo[seqid].color if seqid in chrinfo else defaultcolor
        newseqid = chrinfo[seqid].new_name if seqid in chrinfo else seqid
        if draw_seqids:
            root.text(
                xpos,
                chr_label_y,
                newseqid,
                color=c,
                ha="center",
                va="center",
                rotation=rotation,
            )
        seqid_median = medians[seqid]
        if median_line:
            root.text(
                xpos,
                median_depth_y,
                str(int(seqid_median)),
                color=c,
                ha="center",
                va="center",
            )

    # Plot the regions of interest
    if roi:
        for chrom, pos, name in roi:
            if chrom not in starts:
                continue
            x = starts[chrom] + pos
            # TODO: Remove this special case
            color = {"II": "tomato", "low qual": "g"}.get(name, "gray")
            ax.plot((x, x), (0, maxdepth), "-", lw=2, color=color)

    # Add an arrow to the right of the plot, indicating these are median depths
    if median_line:
        root.text(
            1 - right_margin + 0.01,
            0.88,
            r"$\leftarrow$median",
            color="lightslategray",
            va="center",
        )

    title_pos = 1 - right_margin / 2
    if title:
        root.text(
            title_pos,
            0.5,
            markup(title),
            color="darkslategray",
            ha="center",
            va="center",
            size=15,
        )
    if subtitle:
        root.text(
            title_pos,
            0.3,
            markup(subtitle),
            color="darkslategray",
            ha="center",
            va="center",
            size=15,
        )
    if calculate_coverage:
        cov_pct = percentage(covered_bp, total_bp, precision=0, mode=None)
        root.text(
            title_pos,
            0.1,
            latex(f"cov: {cov_pct}"),
            color="darkslategray",
            ha="center",
            va="center",
            size=15,
        )

    ax.set_xticks([])
    ax.set_xlim(0, xsize)
    if logscale:
        ax.set_yscale("log", basey=2)
    ax.set_ylim(1 if logscale else 0, maxdepth)
    ax.set_ylabel("Depth")

    set_human_axis(ax)
    plt.setp(ax.get_xticklabels() + ax.get_yticklabels(), color="gray", size=12)
    normalize_axes(root)


def read_roi(roi_file: str) -> Dict[str, List[str]]:
    """
    Read the regions of interest file, and return a dict of filename => regions.
    """
    roi = defaultdict(list)
    with open(roi_file, encoding="utf-8") as fp:
        for row in fp:
            filename, region, name = row.strip().split(",")[:3]
            chrom, start_end = region.split(":", 1)
            start, end = start_end.split("-")
            region = (chrom, (int(start) + int(end)) // 2, name)
            roi[filename].append(region)
    logger.info("Read %d regions of interest", len(roi))
    return roi


def draw_multi_depth(
    root,
    panel_roots,
    panel_axes,
    bedfiles: List[str],
    chrinfo_file: str,
    titleinfo_file: str,
    maxdepth: int,
    logscale: bool,
    median_line: bool = True,
    calculate_coverage: bool = False,
    roi: Optional[str] = None,
    left_margin: float = 0.1,
    right_margin: float = 0.15,
):
    """
    Draw multiple depth plots on the same canvas.
    """
    chrinfo = ChrInfoFile(chrinfo_file) if chrinfo_file else {}
    titleinfo = TitleInfoFile(titleinfo_file) if titleinfo_file else {}
    npanels = len(bedfiles)
    yinterval = 1.0 / npanels
    ypos = 1 - yinterval
    roi = read_roi(roi) if roi else {}
    for i, (bedfile, panel_root, panel_ax) in enumerate(
        zip(bedfiles, panel_roots, panel_axes)
    ):
        pf = op.basename(bedfile).split(".", 1)[0]
        bed = Bed(bedfile)

        if ypos > 0.001:
            root.plot((0.02, 0.98), (ypos, ypos), "-", lw=2, color="lightgray")

        title = titleinfo.get(bedfile, pf.split("_", 1)[0])
        subtitle = None
        if isinstance(title, TitleInfoLine):
            subtitle = title.subtitle
            title = title.title

        draw_seqids = i in (0, npanels - 1)
        draw_depth(
            panel_root,
            panel_ax,
            bed,
            chrinfo=chrinfo,
            maxdepth=maxdepth,
            logscale=logscale,
            title=title,
            subtitle=subtitle,
            median_line=median_line,
            draw_seqids=draw_seqids,
            calculate_coverage=calculate_coverage,
            roi=roi.get(bedfile),
            left_margin=left_margin,
            right_margin=right_margin,
        )
        ypos -= yinterval

    normalize_axes(root)


def depth(args):
    """
    %prog depth *.regions.bed.gz

    Plot the mosdepth regions BED file. We recommend to generate this BED file
    by (please adjust the --by parameter to your required resolution):

    $ mosdepth --no-per-base --use-median --fast-mode --by 1000000 sample.wgs
    sample.bam

    Use --chrinfo to specify a colormap between seqid, desired color, and
    optionally a new name. For example:

    chr01A, #c51b7d, 1A
    chr01B, #4d9221, 1B
    ...

    Only seqids that are in the colormap will be plotted, in the order that's
    given in the file. When --colormap is not set, every seqid will be drawn in
    black.

    Can take multiple BED files as input and then plot all of them in a
    composite figure.
    """
    p = OptionParser(depth.__doc__)
    p.add_argument(
        "--chrinfo", help="Comma-separated mappings between seqid, color, new_name"
    )
    p.add_argument(
        "--titleinfo",
        help="Comma-separated titles mappings between filename, title",
    )
    p.add_argument("--maxdepth", default=100, type=int, help="Maximum depth to show")
    p.add_argument(
        "--logscale", default=False, action="store_true", help="Use log-scale on depth"
    )
    p.add_argument(
        "--no-median-line",
        default=False,
        action="store_true",
        help="Do not plot median depth line",
    )
    p.add_argument(
        "--calculate-coverage",
        default=False,
        action="store_true",
        help="Calculate genome coverage",
    )
    p.add_argument(
        "--roi",
        help="File that contains regions of interest, format: filename, chr:start-end",
    )
    p.add_argument("--left-margin", default=0.1, type=float, help="Left margin")
    p.add_argument("--right-margin", default=0.15, type=float, help="Right margin")
    p.set_outfile("depth.pdf")
    opts, args, iopts = p.set_image_options(args, style="dark", figsize="10x8")

    if len(args) < 1:
        sys.exit(not p.print_help())

    bedfiles = args

    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes((0, 0, 1, 1))

    npanels = len(bedfiles)
    yinterval = 1.0 / npanels
    ypos = 1 - yinterval
    panel_roots, panel_axes = [], []
    left_margin, right_margin = opts.left_margin, opts.right_margin
    width = 1 - left_margin - right_margin
    for _ in range(npanels):
        panel_root = root if npanels == 1 else fig.add_axes((0, ypos, 1, yinterval))
        panel_ax = fig.add_axes(
            (left_margin, ypos + 0.2 * yinterval, width, 0.65 * yinterval)
        )
        panel_roots.append(panel_root)
        panel_axes.append(panel_ax)
        ypos -= yinterval

    draw_multi_depth(
        root,
        panel_roots,
        panel_axes,
        bedfiles,
        opts.chrinfo,
        opts.titleinfo,
        opts.maxdepth,
        opts.logscale,
        median_line=not opts.no_median_line,
        calculate_coverage=opts.calculate_coverage,
        roi=opts.roi,
        left_margin=left_margin,
        right_margin=right_margin,
    )

    image_name = opts.outfile
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)
    return image_name


def add_window_options(p):
    """
    Add options for window plotting.
    """
    p.add_argument("--window", default=500000, type=int, help="Size of window")
    p.add_argument("--shift", default=100000, type=int, help="Size of shift")
    p.add_argument("--subtract", help="Subtract bases from window")
    p.add_argument(
        "--nomerge", default=False, action="store_true", help="Do not merge features"
    )


def check_window_options(opts):
    """
    Check the window options, and return the values.
    """
    window = opts.window
    shift = opts.shift
    subtract = opts.subtract
    assert window % shift == 0, "--window must be divisible by --shift"
    logger.debug(
        "Line/stack-plot options: window=%d shift=%d subtract=%s",
        window,
        shift,
        subtract,
    )
    merge = not opts.nomerge

    return window, shift, subtract, merge


def get_beds(s: List[str], binned: bool = False) -> List[str]:
    """
    Get the bed files for each feature, and return them as a list.
    """
    return [x + ".bed" for x in s] if not binned else [x for x in s]


def linearray(binfile, chr, window, shift):
    mn = binfile.mapping[chr]
    m, _ = zip(*mn)

    m = np.array(m, dtype=float)
    w = window // shift
    m = moving_sum(m, window=w)
    return m


def lineplot(ax, binfiles, nbins, chr, window, shift, color="br"):
    assert len(binfiles) <= 2, "A max of two line plots are supported"

    t = np.arange(nbins)
    bf = binfiles[0]
    m = linearray(bf, chr, window, shift)
    ax.plot(t, m, "{0}-".format(color[0]), lw=2)

    formatter = ticker.FuncFormatter(
        lambda x, pos: human_readable_base(int(x) * shift, pos)
    )
    ax.xaxis.set_major_formatter(formatter)
    for tl in ax.get_xticklabels():
        tl.set_color("darkslategray")

    label = bf.filename.split(".")[0]
    perw = "per {0}".format(human_size(window, precision=0))
    ax.set_ylabel(label + " " + perw, color=color[0])

    if len(binfiles) == 2:
        ax2 = ax.twinx()
        bf = binfiles[1]
        m = linearray(bf, chr, window, shift)
        ax2.plot(t, m, "{0}-".format(color[1]), lw=2)
        # Differentiate tick labels through colors
        for tl in ax.get_yticklabels():
            tl.set_color(color[0])
        for tl in ax2.get_yticklabels():
            tl.set_color(color[1])

        label = bf.filename.split(".")[0]
        ax2.set_ylabel(label + " " + perw, color=color[1])

    ax.set_xlim(0, nbins)


def composite(args):
    """
    %prog composite fastafile chr1

    Combine line plots, feature bars and alt-bars, different data types
    specified in options. Inputs must be BED-formatted. Three types of viz are
    currently supported:

    --lines: traditional line plots, useful for plotting feature freq
    --bars: show where the extent of features are
    --altbars: similar to bars, yet in two alternating tracks, e.g. scaffolds
    """
    p = OptionParser(composite.__doc__)
    p.add_argument("--lines", help="Features to plot in lineplot")
    p.add_argument("--bars", help="Features to plot in bars")
    p.add_argument("--altbars", help="Features to plot in alt-bars")
    p.add_argument(
        "--fatten",
        default=False,
        action="store_true",
        help="Help visualize certain narrow features",
    )
    p.add_argument(
        "--mode",
        default="span",
        choices=("span", "count", "score"),
        help="Accumulate feature based on",
    )
    add_window_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="8x5")

    if len(args) != 2:
        sys.exit(not p.print_help())

    fastafile, chr = args
    window, shift, _, merge = check_window_options(opts)
    linebeds, barbeds, altbarbeds = [], [], []
    fatten = opts.fatten
    if opts.lines:
        lines = opts.lines.split(",")
        linebeds = get_beds(lines)
    if opts.bars:
        bars = opts.bars.split(",")
        barbeds = get_beds(bars)
    if opts.altbars:
        altbars = opts.altbars.split(",")
        altbarbeds = get_beds(altbars)

    linebins = get_binfiles(linebeds, fastafile, shift, mode=opts.mode, merge=merge)

    margin = 0.12
    clen = Sizes(fastafile).mapping[chr]
    nbins, _ = get_nbins(clen, shift)

    plt.rcParams["xtick.major.size"] = 0
    plt.rcParams["ytick.major.size"] = 0

    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes((0, 0, 1, 1))

    root.text(0.5, 0.95, chr, ha="center", color="darkslategray")

    xstart, xend = margin, 1 - margin
    xlen = xend - xstart
    ratio = xlen / clen
    # Line plots
    ax = fig.add_axes((xstart, 0.6, xlen, 0.3))
    lineplot(ax, linebins, nbins, chr, window, shift)

    # Bar plots
    yy = 0.5
    yinterval = 0.08
    xs = lambda x: xstart + ratio * x
    r = 0.01
    fattend = 0.0025
    for bb in barbeds:
        root.text(xend + 0.01, yy, bb.split(".")[0], va="center")
        HorizontalChromosome(root, xstart, xend, yy, height=0.02)
        bb = Bed(bb)
        for b in bb:
            start, end = xs(b.start), xs(b.end)
            span = end - start
            if fatten and span < fattend:
                span = fattend

            root.add_patch(
                Rectangle((start, yy - r), span, 2 * r, lw=0, fc="darkslategray")
            )
        yy -= yinterval

    # Alternative bar plots
    offset = r / 2
    for bb in altbarbeds:
        root.text(xend + 0.01, yy, bb.split(".")[0], va="center")
        bb = Bed(bb)
        for b in bb:
            start, end = xs(b.start), xs(b.end)
            span = end - start
            if span < 0.0001:
                continue
            offset = -offset
            root.add_patch(
                Rectangle(
                    (start, yy + offset), end - start, 0.003, lw=0, fc="darkslategray"
                )
            )
        yy -= yinterval

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    image_name = chr + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def multilineplot(args):
    """
    %prog multilineplot fastafile chr1

    Combine multiple line plots in one vertical stack
    Inputs must be BED-formatted.

    --lines: traditional line plots, useful for plotting feature freq
    """
    p = OptionParser(multilineplot.__doc__)
    p.add_argument("--lines", help="Features to plot in lineplot")
    p.add_argument("--colors", help="List of colors matching number of input bed files")
    p.add_argument(
        "--mode",
        default="span",
        choices=("span", "count", "score"),
        help="Accumulate feature based on",
    )
    p.add_argument(
        "--binned",
        default=False,
        action="store_true",
        help="Specify whether the input is already binned; "
        + "if True, input files are considered to be binfiles",
    )
    p.add_argument("--ymax", type=int, help="Set Y-axis max")
    add_window_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="8x5")

    if len(args) != 2:
        sys.exit(not p.print_help())

    fastafile, chr = args
    window, shift, _, merge = check_window_options(opts)
    linebeds = []
    colors = opts.colors
    if opts.lines:
        lines = opts.lines.split(",")
        assert len(colors) == len(lines), (
            "Number of chosen colors must match" + " number of input bed files"
        )
        linebeds = get_beds(lines, binned=opts.binned)

    linebins = get_binfiles(
        linebeds, fastafile, shift, mode=opts.mode, binned=opts.binned, merge=merge
    )

    clen = Sizes(fastafile).mapping[chr]
    nbins, _ = get_nbins(clen, shift)

    plt.rcParams["xtick.major.size"] = 0
    plt.rcParams["ytick.major.size"] = 0
    plt.rcParams["figure.figsize"] = iopts.w, iopts.h

    fig, axarr = plt.subplots(nrows=len(lines))
    if len(linebeds) == 1:
        axarr = (axarr,)
    fig.suptitle(latex(chr), color="darkslategray")

    for i, ax in enumerate(axarr):
        lineplot(
            ax,
            [linebins[i]],
            nbins,
            chr,
            window,
            shift,
            color="{0}{1}".format(colors[i], "r"),
        )

    if opts.ymax:
        ax.set_ylim(0, opts.ymax)

    plt.subplots_adjust(hspace=0.5)

    image_name = chr + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def draw_heatmaps(
    fig,
    root,
    root_extent: Extent,
    fastafile: str,
    chr: str,
    stacks: List[str],
    heatmaps: List[str],
    window: int,
    shift: int,
    cmap: Colormap,
    subtract: Optional[int] = None,
    merge: bool = False,
    meres: Optional[str] = None,
):
    """
    Draw heatmap for the given chromosome.
    """
    stackbeds = get_beds(stacks)
    heatmapbeds = get_beds(heatmaps)
    stackbins = get_binfiles(
        stackbeds, fastafile, shift, subtract=subtract, merge=merge
    )
    heatmapbins = get_binfiles(
        heatmapbeds, fastafile, shift, subtract=subtract, merge=merge
    )

    margin = 0.06
    inner = 0.015
    clen = Sizes(fastafile).mapping[chr]

    # Gauge
    ratio = draw_gauge(root, margin, clen, rightmargin=4 * margin)
    yinterval = 0.3
    xx = margin
    yy = 1 - margin
    yy -= yinterval
    xlen = clen / ratio
    cc = chr
    if "_" in chr:
        ca, cb = chr.split("_")
        cc = ca[0].upper() + cb

    root.add_patch(Rectangle((xx, yy), xlen, yinterval - inner, color=gray))
    extent = (xx, yy, xlen, yinterval - inner)
    adjusted = adjust_extent(extent, root_extent)
    ax = fig.add_axes(adjusted)

    nbins, _ = get_nbins(clen, shift)

    owindow = clen / 100
    if owindow > window:
        window = owindow // shift * shift

    stackplot(ax, stackbins, nbins, palette, chr, window, shift)
    ax.text(
        0.05,
        0.9,
        cc,
        va="top",
        zorder=100,
        transform=ax.transAxes,
        bbox=dict(boxstyle="round", fc="w", alpha=0.5),
    )

    # Legends
    xx += xlen + 0.01
    yspace = (yinterval - inner) / (len(stackbins) + 1)
    yy = 1 - margin - yinterval
    for s, p in zip(stacks, palette):
        s = s.replace("_", " ")
        s = Registration.get(s, s)

        yy += yspace
        root.add_patch(Rectangle((xx, yy), inner, inner, color=p, lw=0))
        root.text(xx + 1.5 * inner, yy, s, size=10)

    yh = 0.05  # Heatmap height
    # Heatmaps
    xx = margin
    yy = 1 - margin - yinterval - inner
    for s, p in zip(heatmaps, heatmapbins):
        s = s.replace("_", " ")
        s = Registration.get(s, s)

        yy -= yh
        m = stackarray(p, chr, window, shift)

        Y = np.array([m, m])
        root.imshow(
            Y,
            extent=(xx, xx + xlen, yy, yy + yh - inner),
            interpolation="nearest",
            aspect="auto",
            cmap=cmap,
        )
        root.text(xx + xlen + 0.01, yy, s, size=10)

    yy -= yh

    if meres:
        bed = Bed(meres)
        for b in bed:
            if b.seqid != chr:
                continue
            pos = (b.start + b.end) / 2
            cpos = pos / ratio
            xx = margin + cpos
            accn = b.accn.capitalize()
            root.add_patch(CirclePolygon((xx, yy), radius=0.01, fc="m", ec="m"))
            root.text(xx + 0.014, yy, accn, va="center", color="m")

    normalize_axes(root)


def heatmap(args):
    """
    %prog heatmap fastafile chr1

    Combine stack plot with heatmap to show abundance of various tracks along
    given chromosome. Need to give multiple beds to --stacks and --heatmaps
    """
    p = OptionParser(heatmap.__doc__)
    p.add_argument(
        "--stacks",
        default="Exons,Introns,DNA_transposons,Retrotransposons",
        help="Features to plot in stackplot",
    )
    p.add_argument(
        "--heatmaps",
        default="Copia,Gypsy,hAT,Helitron,Introns,Exons",
        help="Features to plot in heatmaps",
    )
    p.add_argument("--meres", default=None, help="Extra centromere / telomere features")
    add_window_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="8x5")

    if len(args) != 2:
        sys.exit(not p.print_help())

    fastafile, chr = args
    window, shift, subtract, merge = check_window_options(opts)

    stacks = opts.stacks.split(",")
    heatmaps = opts.heatmaps.split(",")

    fig = plt.figure(1, (iopts.w, iopts.h))
    root_extent = (0, 0, 1, 1)
    root = fig.add_axes(root_extent)

    draw_heatmaps(
        fig,
        root,
        root_extent,
        fastafile,
        chr,
        stacks,
        heatmaps,
        window,
        shift,
        iopts.cmap,
        subtract,
        merge,
        meres=opts.meres,
    )

    image_name = chr + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def draw_gauge(ax, margin: float, maxl: int, rightmargin: Optional[float] = None):
    """
    Draw a gauge on the top of the canvas, showing the scale of the chromosome.
    """
    rightmargin = rightmargin or margin
    ax.plot([margin, 1 - rightmargin], [1 - margin, 1 - margin], "k-", lw=2)

    best_stride = autoscale(maxl)
    nintervals = maxl / best_stride

    xx, yy = margin, 1 - margin
    tip = 0.005
    xinterval = (1 - margin - rightmargin) / nintervals
    l = human_size(best_stride)
    if l[-1] == "b":
        suffix = target = l[-2:]

    for i in range(0, maxl + 1, best_stride):
        l = human_size(i, precision=0, target=target)
        if l[-1] == "b":
            l, suffix = l[:-2], l[-2:]
        ax.plot([xx, xx], [yy, yy + tip], "k-", lw=2)
        ax.text(xx, yy + 2 * tip, l, ha="center", size=13)
        xx += xinterval

    xx += 4 * tip - xinterval
    ax.text(xx + tip, yy + 2 * tip, suffix)

    return best_stride / xinterval


def get_binfiles(
    inputfiles: List[str],
    fastafile: str,
    shift: int,
    mode: str = "span",
    subtract: Optional[int] = None,
    binned: bool = False,
    merge: bool = True,
):
    """
    Get binfiles from input files. If not binned, then bin them first.
    """
    if not binned:
        binopts = [f"--binsize={shift}"]
        binopts.append(f"--mode={mode}")
        if subtract:
            binopts.append(f"--subtract={subtract}")
        if not merge:
            binopts.append("--nomerge")
        binfiles = [bins([x, fastafile] + binopts) for x in inputfiles if op.exists(x)]
    else:
        binfiles = inputfiles
    binfiles = [BinFile(x) for x in binfiles]

    return binfiles


def stackarray(binfile: BinFile, chr: str, window: int, shift: int):
    """
    Get stack array from binfile for the given chr.
    """
    mn = binfile.mapping[chr]
    m, n = zip(*mn)

    m = np.array(m, dtype=float)
    n = np.array(n, dtype=float)

    w = window // shift
    nw = m.shape[0]
    if nw < w:
        logger.info("%s length < window, using %d bins instead of %d", chr, nw, w)
        w = nw
    m = moving_sum(m, window=w)
    n = moving_sum(n, window=w)
    m /= n

    return m


def stackplot(
    ax,
    binfiles: List[BinFile],
    nbins: int,
    palette: List[str],
    chr: str,
    window: int,
    shift: int,
):
    """
    Plot stackplot on the given axes, using data from binfiles.
    """
    t = np.arange(nbins, dtype=float) + 0.5
    m = np.zeros(nbins, dtype=float)
    zorders = range(10)[::-1]
    for binfile, p, z in zip(binfiles, palette, zorders):
        s = stackarray(binfile, chr, window, shift)
        m += s
        ax.fill_between(t, m, color=p, lw=0, zorder=z)

    ax.set_xlim(0, nbins)
    ax.set_ylim(0, 1)
    ax.set_axis_off()


def draw_stacks(
    fig,
    root,
    root_extent: Extent,
    stacks: List[str],
    fastafile: str,
    window: int,
    shift: int,
    top: int,
    merge: bool = True,
    subtract: Optional[int] = None,
    switch: Optional[DictFile] = None,
):
    """
    Draw stack plot.
    """
    bedfiles = get_beds(stacks)
    binfiles = get_binfiles(bedfiles, fastafile, shift, subtract=subtract, merge=merge)

    sizes = Sizes(fastafile)
    s = list(sizes.iter_sizes())[:top]
    maxl = max(x[1] for x in s)
    margin = 0.08
    inner = 0.02  # y distance between tracks

    # Gauge
    ratio = draw_gauge(root, margin, maxl)

    # Per chromosome
    yinterval = (1 - 2 * margin) / (top + 1)
    xx = margin
    yy = 1 - margin
    for chr, clen in s:
        yy -= yinterval
        xlen = clen / ratio
        cc = chr
        if "_" in chr:
            ca, cb = chr.split("_")
            cc = ca[0].upper() + cb

        if switch and cc in switch:
            cc = "\n".join((cc, f"({switch[cc]})"))

        extent = (xx, yy, xlen, yinterval - inner)
        adjusted = adjust_extent(extent, root_extent)
        root.add_patch(Rectangle((xx, yy), xlen, yinterval - inner, color=gray))
        ax = fig.add_axes(adjusted)

        nbins, _ = get_nbins(clen, shift)

        stackplot(ax, binfiles, nbins, palette, chr, window, shift)
        root.text(
            xx - 0.04, yy + 0.5 * (yinterval - inner), cc, ha="center", va="center"
        )

    # Legends
    yy -= yinterval
    xx = margin
    for b, p in zip(bedfiles, palette):
        b = b.rsplit(".", 1)[0].replace("_", " ")
        b = Registration.get(b, b)

        root.add_patch(Rectangle((xx, yy), inner, inner, color=p, lw=0))
        xx += 2 * inner
        root.text(xx, yy, b, size=13)
        xx += len(b) * 0.015 + inner

    normalize_axes(root)


def stack(args):
    """
    %prog stack fastafile

    Create landscape plots that show the amounts of genic sequences, and repetitive
    sequences along the chromosomes.
    """
    p = OptionParser(stack.__doc__)
    p.add_argument("--top", default=10, type=int, help="Draw the first N chromosomes")
    p.add_argument(
        "--stacks",
        default="Exons,Introns,DNA_transposons,Retrotransposons",
        help="Features to plot in stackplot",
    )
    p.add_argument("--switch", help="Change chr names based on two-column file")
    add_window_options(p)
    opts, args, iopts = p.set_image_options(args, figsize="8x8")

    if len(args) != 1:
        sys.exit(not p.print_help())

    (fastafile,) = args
    top = opts.top
    window, shift, subtract, merge = check_window_options(opts)
    switch = opts.switch
    if switch:
        switch = DictFile(opts.switch)

    stacks = opts.stacks.split(",")

    fig = plt.figure(1, (iopts.w, iopts.h))
    root_extent = (0, 0, 1, 1)
    root = fig.add_axes(root_extent)

    draw_stacks(
        fig,
        root,
        root_extent,
        stacks,
        fastafile,
        window,
        shift,
        top,
        merge,
        subtract,
        switch,
    )

    pf = fastafile.rsplit(".", 1)[0]
    image_name = pf + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)
    return image_name


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Assortment of utility scripts implementing recipes from Python cookbooks, such as iterators, sorters, range queries, etc.
"""

from ..apps.base import dmain

if __name__ == "__main__":
    dmain(__file__)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Miscellany of wrapper scripts for command-line bioinformatics tools, public data downloaders and other generic routines.
"""

from .base import dmain

if __name__ == "__main__":
    dmain(__file__)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
MAF format specification:
<http://genome.ucsc.edu/FAQ/FAQformat#format5>
"""
from bisect import bisect
from dataclasses import dataclass
import sys

from Bio import AlignIO, SeqIO
from bx import interval_index_file
from bx.align import maf

from ..apps.base import ActionDispatcher, OptionParser, need_update
from ..apps.lastz import blastz_score_to_ncbi_bits, blastz_score_to_ncbi_expectation
from .base import BaseFile, logger

FLANK = 60


class Maf(BaseFile, dict):
    def __init__(self, filename, index=False):
        super().__init__(filename)

        indexfile = filename + ".idx"
        if index:
            if need_update(filename, indexfile):
                self.build_index(filename, indexfile)

            self.index = maf.Index(filename, indexfile)

        fp = open(filename)
        self.reader = maf.Reader(fp)

    def build_index(self, filename, indexfile):
        """
        Recipe from Brad Chapman's blog
        <http://bcbio.wordpress.com/2009/07/26/sorting-genomic-alignments-using-python/>
        """
        indexes = interval_index_file.Indexes()
        in_handle = open(filename)

        reader = maf.Reader(in_handle)
        while True:
            pos = reader.file.tell()
            rec = next(reader)
            if rec is None:
                break
            for c in rec.components:
                indexes.add(
                    c.src,
                    c.forward_strand_start,
                    c.forward_strand_end,
                    pos,
                    max=c.src_size,
                )

        index_handle = open(indexfile, "w")
        indexes.write(index_handle)
        index_handle.close()


@dataclass
class Breakpoint:
    arec: str
    astart: int
    brec: str
    bstart: int

    def __str__(self):
        return f"{self.arec}:{self.astart}-{self.brec}:{self.bstart}"


def main():

    actions = (
        ("bed", "convert MAF to BED format"),
        ("blast", "convert MAF to BLAST tabular format"),
        ("breakpoints", "find breakpoints in MAF and 'simulate' chimeric contigs"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def breakpoints(args):
    """
    %prog breakpoints A.B.maf A.fa B.fa AB 1000000 2000000

    Find breakpoints in MAF and 'simulate' chimeric contigs in `AB.fa`.
    Breakpoints are 'roughly' around the user defined positions. The idea is
    to simulate chimeric contigs, which are useful for testing algorithms,
    e.g. klassify.
    """
    p = OptionParser(breakpoints.__doc__)
    p.add_argument(
        "--minsize",
        default=10000,
        type=int,
        help="Minimum size of alignment to consider",
    )
    opts, args = p.parse_args(args)

    if len(args) not in (5, 6):
        sys.exit(not p.print_help())

    maf_file, a_fasta, b_fasta, ab = args[:4]
    bps = sorted(int(x) for x in args[4:])
    minsize = opts.minsize

    filtered_msa = []
    for msa in AlignIO.parse(maf_file, "maf"):
        arec, brec = msa
        if brec.annotations["size"] < minsize:
            continue
        filtered_msa.append((brec.annotations["start"], arec, brec))
    logger.info("Total alignments: %d", len(filtered_msa))

    final = []
    # Load the sequences
    ar = next(SeqIO.parse(a_fasta, "fasta"))
    br = next(SeqIO.parse(b_fasta, "fasta"))
    for bp in bps:
        i = bisect(filtered_msa, (bp,))
        _, arec, brec = filtered_msa[i]
        logger.info("%s", arec)
        logger.info("%s", brec)
        assert len(arec) == len(brec)
        # Find the midpoint, safe to crossover there
        midpoint = len(arec) // 2
        aseq = arec.seq[:midpoint]
        astart = arec.annotations["start"] + len(aseq) - aseq.count("-")
        logger.info("%s|%s", aseq[-FLANK:], arec.seq[midpoint:][:FLANK])
        bseq = brec.seq[:midpoint]
        bstart = brec.annotations["start"] + len(bseq) - bseq.count("-")
        logger.info("%s|%s", bseq[-FLANK:], brec.seq[midpoint:][:FLANK])
        bpt = Breakpoint(arec.id, astart, brec.id, bstart)
        logger.info("-" * FLANK * 2 + ">")
        logger.info("%s|%s", ar.seq[:astart][-FLANK:], br.seq[bstart:][:FLANK])
        final.append(bpt)

    logger.info("Breakpoints found: %s", final)
    if len(final) == 2:
        bp1, bp2 = final[:2]
        # ====-------=======
        #   bp1     bp2
        abseq = (
            ar.seq[: bp1.astart]
            + br.seq[bp1.bstart : bp2.bstart]
            + ar.seq[bp2.astart :]
        )
    elif len(final) == 1:
        bp = final[0]
        abseq = ar.seq[: bp.astart] + br.seq[bp.bstart :]
    abrec = SeqIO.SeqRecord(abseq, id=ab, description="")
    ab_fasta = f"{ab}.fa"
    SeqIO.write([abrec], ab_fasta, "fasta")
    logger.info("Writing to %s", ab_fasta)


def bed(args):
    """
    %prog bed maffiles > out.bed

    Convert a folder of maf alignments to the bed features
    then useful to check coverage, etc.
    """
    p = OptionParser(bed.__doc__)
    _, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(p.print_help())

    flist = args
    prefix = flist[0].split(".")[0]

    j = 0
    for f in flist:
        reader = Maf(f).reader
        for rec in reader:
            a, b = rec.components

            for a, tag in zip((a, b), "ab"):
                name = "{0}_{1:07d}{2}".format(prefix, j, tag)
                print(
                    "\t".join(
                        str(x)
                        for x in (
                            a.src,
                            a.forward_strand_start,
                            a.forward_strand_end,
                            name,
                        )
                    )
                )

            j += 1


def alignment_details(a, b):
    nmatch = 0
    nmismatch = 0
    ngaps = 0

    assert len(a) == len(b)
    l = len(a)

    for i in range(l):
        if a[i] == b[i]:
            nmatch += 1
        elif a[i] == "-" or b[i] == "-":
            ngaps += 1
        else:
            nmismatch += 1

    pctid = 100.0 * nmatch / l
    return pctid, nmismatch, ngaps


def maf_to_blast8(f):
    """
    Convert a MAF file to BLAST tabular format.
    """
    reader = Maf(f).reader
    for rec in reader:
        a, b = rec.components
        query = a.src
        subject = b.src
        qstart = a.forward_strand_start
        qstop = a.forward_strand_end
        sstart = b.forward_strand_start
        sstop = b.forward_strand_end
        score = rec.score

        evalue = blastz_score_to_ncbi_expectation(score)
        score = blastz_score_to_ncbi_bits(score)
        evalue, score = "{0:.2g}".format(evalue), "{0:.1f}".format(score)
        hitlen = len(a.text)

        pctid, nmismatch, ngaps = alignment_details(a.text, b.text)
        print(
            "\t".join(
                str(x)
                for x in (
                    query,
                    subject,
                    pctid,
                    hitlen,
                    nmismatch,
                    ngaps,
                    qstart,
                    qstop,
                    sstart,
                    sstop,
                    evalue,
                    score,
                )
            )
        )


def blast(args):
    """
    %prog blast maffiles > out.blast

    From a folder of .maf files, generate .blast file with tabular format.
    """
    p = OptionParser(blast.__doc__)
    _, args = p.parse_args(args)

    if len(args) == 0:
        sys.exit(p.print_help())

    flist = args

    for f in flist:
        maf_to_blast8(f)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Scripts for the Brapa bites paper

Tang et al. (2012) Altered Patterns of Fractionation and Exon Deletions in
Brassica rapa Support a Two-Step Model of Paleohexaploidy. Genetics.
<http://www.genetics.org/content/190/4/1563.short>
"""
from more_itertools import pairwise

from ..apps.base import ActionDispatcher, OptionParser, fname
from ..graphics.base import CirclePolygon, Polygon, Rectangle, plt, savefig
from ..graphics.chromosome import Chromosome
from ..graphics.glyph import RoundLabel, TextCircle, arrowprops


def main():

    actions = (
        ("excision", "show intra-chromosomal recombination"),
        ("bites", "show the bites calling pipeline"),
        ("scenario", "show step-wise genome merger events in brapa"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def excision(args):
    """
    %prog excision

    Illustrate the mechanism of illegitimate recombination.
    """
    p = OptionParser(__doc__)
    opts, args = p.parse_args(args)

    fig = plt.figure(1, (5, 5))
    root = fig.add_axes([0, 0, 1, 1])

    plt.plot((0.2, 0.8), (0.6, 0.6), "r-", lw=3)
    plt.plot((0.4, 0.6), (0.6, 0.6), "b>-", mfc="g", mec="w", ms=12, lw=3)
    plt.plot((0.3, 0.7), (0.5, 0.5), "r-", lw=3)
    plt.plot((0.5,), (0.5,), "b>-", mfc="g", mec="w", ms=12, lw=3)

    # Circle excision
    plt.plot((0.5,), (0.45,), "b>-", mfc="g", mec="w", ms=12, lw=3)
    circle = CirclePolygon((0.5, 0.4), 0.05, fill=False, lw=3, ec="b")
    root.add_patch(circle)

    arrow_dist = 0.07
    ar_xpos, ar_ypos = 0.5, 0.52
    root.annotate(
        " ", (ar_xpos, ar_ypos), (ar_xpos, ar_ypos + arrow_dist), arrowprops=arrowprops
    )

    RoundLabel(root, 0.2, 0.64, "Gene")
    RoundLabel(root, 0.3, 0.54, "Excision")

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    figname = fname() + ".pdf"
    savefig(figname, dpi=300)


def bites(args):
    """
    %prog bites

    Illustrate the pipeline for automated bite discovery.
    """

    p = OptionParser(__doc__)
    opts, args = p.parse_args()

    fig = plt.figure(1, (6, 6))
    root = fig.add_axes([0, 0, 1, 1])

    # HSP pairs
    hsps = (
        ((50, 150), (60, 180)),
        ((190, 250), (160, 235)),
        ((300, 360), (270, 330)),
        ((430, 470), (450, 490)),
        ((570, 620), (493, 543)),
        ((540, 555), (370, 385)),  # non-collinear hsps
    )

    titlepos = (0.9, 0.65, 0.4)
    titles = ("Compare orthologous region", "Find collinear HSPs", "Scan paired gaps")
    ytip = 0.01
    mrange = 650.0
    m = lambda x: x / mrange * 0.7 + 0.1
    for i, (ya, title) in enumerate(zip(titlepos, titles)):
        yb = ya - 0.1
        plt.plot((0.1, 0.8), (ya, ya), "-", color="gray", lw=2, zorder=1)
        plt.plot((0.1, 0.8), (yb, yb), "-", color="gray", lw=2, zorder=1)
        RoundLabel(root, 0.5, ya + 4 * ytip, title)
        root.text(0.9, ya, "A. thaliana", ha="center", va="center")
        root.text(0.9, yb, "B. rapa", ha="center", va="center")
        myhsps = hsps
        if i >= 1:
            myhsps = hsps[:-1]
        for (a, b), (c, d) in myhsps:
            a, b, c, d = [m(x) for x in (a, b, c, d)]
            r1 = Rectangle((a, ya - ytip), b - a, 2 * ytip, fc="r", lw=0, zorder=2)
            r2 = Rectangle((c, yb - ytip), d - c, 2 * ytip, fc="r", lw=0, zorder=2)
            r3 = Rectangle((a, ya - ytip), b - a, 2 * ytip, fill=False, zorder=3)
            r4 = Rectangle((c, yb - ytip), d - c, 2 * ytip, fill=False, zorder=3)
            r5 = Polygon(
                ((a, ya - ytip), (c, yb + ytip), (d, yb + ytip), (b, ya - ytip)),
                fc="r",
                alpha=0.2,
            )
            rr = (r1, r2, r3, r4, r5)
            if i == 2:
                rr = rr[:-1]
            for r in rr:
                root.add_patch(r)

    # Gap pairs
    hspa, hspb = zip(*myhsps)
    gapa, gapb = [], []
    for (a, b), (c, d) in pairwise(hspa):
        gapa.append((b + 1, c - 1))
    for (a, b), (c, d) in pairwise(hspb):
        gapb.append((b + 1, c - 1))
    gaps = zip(gapa, gapb)
    tpos = titlepos[-1]

    yy = tpos - 0.05
    for i, ((a, b), (c, d)) in enumerate(gaps):
        i += 1
        a, b, c, d = [m(x) for x in (a, b, c, d)]
        xx = (a + b + c + d) / 4
        TextCircle(root, xx, yy, str(i))

    # Bites
    ystart = 0.24
    ytip = 0.05
    bites = (
        ("Bite(40=>-15)", True),
        ("Bite(50=>35)", False),
        ("Bite(70=>120)", False),
        ("Bite(100=>3)", True),
    )
    for i, (bite, selected) in enumerate(bites):
        xx = 0.15 if (i % 2 == 0) else 0.55
        yy = ystart - i / 2 * ytip
        i += 1
        TextCircle(root, xx, yy, str(i))
        color = "k" if selected else "gray"
        root.text(xx + ytip, yy, bite, size=10, color=color, va="center")

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    figname = fname() + ".pdf"
    savefig(figname, dpi=300)


def scenario(args):
    """
    %prog scenario

    Illustration of the two-step genome merger process for B. rapa companion paper.
    """
    p = OptionParser(__doc__)
    p.parse_args()

    fig = plt.figure(1, (5, 5))
    root = fig.add_axes([0, 0, 1, 1])

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    # Layout format: (x, y, label, (chr lengths))
    anc = (0.5, 0.9, "Ancestor", (1,))
    s1 = (0.2, 0.6, "Genome I", (1,))
    s2 = (0.5, 0.6, "Genome II", (1,))
    s3 = (0.8, 0.6, "Genome III", (1,))
    tetra = (0.35, 0.4, "Tetraploid I / II", (0.5, 0.9))
    hexa = (0.5, 0.1, "Hexaploid I / II / III", (0.36, 0.46, 0.9))
    labels = (anc, s1, s2, s3, tetra, hexa)
    connections = (
        (anc, s1),
        (anc, s2),
        (anc, s3),
        (s1, tetra),
        (s2, tetra),
        (tetra, hexa),
        (s3, hexa),
    )

    xinterval = 0.02
    yratio = 0.05
    for xx, yy, label, chrl in labels:
        # RoundLabel(root, xx, yy, label)
        root.text(xx, yy, label, ha="center", va="center")
        offset = len(label) * 0.012
        for i, c in enumerate(chrl):
            ya = yy + yratio * c
            yb = yy - yratio * c
            Chromosome(root, xx - offset + i * xinterval, ya, yb, width=0.01)

    # Comments
    comments = ((0.15, 0.33, "II dominant"), (0.25, 0.03, "III dominant"))

    for xx, yy, c in comments:
        root.text(xx, yy, c, size=9, ha="center", va="center")

    # Branches
    tip = 0.04
    for a, b in connections:
        xa, ya, la, chra = a
        xb, yb, lb, chrb = b
        plt.plot((xa, xb), (ya - tip, yb + 2 * tip), "k-", lw=2, alpha=0.5)

    figname = fname() + ".pdf"
    savefig(figname, dpi=300)


if __name__ == "__main__":
    main()
